from collections import defaultdict
from dataclasses import dataclass
from typing import List, Optional

from trilogy.core.enums import (
    BooleanOperator,
    Derivation,
    JoinType,
    SourceType,
)
from trilogy.core.models.build import (
    BuildComparison,
    BuildConcept,
    BuildConditional,
    BuildDatasource,
    BuildGrain,
    BuildOrderBy,
    BuildParenthetical,
    LooseBuildConceptList,
)
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.models.execute import ConceptPair, QueryDatasource, UnnestJoin
from trilogy.utility import unique


def resolve_concept_map(
    inputs: List[QueryDatasource | BuildDatasource],
    targets: List[BuildConcept],
    inherited_inputs: List[BuildConcept],
    full_joins: List[BuildConcept] | None = None,
) -> dict[str, set[BuildDatasource | QueryDatasource | UnnestJoin]]:

    targets = targets or []
    concept_map: dict[str, set[BuildDatasource | QueryDatasource | UnnestJoin]] = (
        defaultdict(set)
    )
    full_addresses = {c.address for c in full_joins} if full_joins else set()
    inherited = set([t.address for t in inherited_inputs])
    for input in inputs:
        for concept in input.output_concepts:
            if concept.address not in input.full_concepts:
                continue
            if (
                isinstance(input, QueryDatasource)
                and concept.address in input.hidden_concepts
            ):
                continue
            if concept.address in full_addresses:
                concept_map[concept.address].add(input)
            elif concept.address not in concept_map:
                # equi_targets = [x for x in targets if concept.address in x.pseudonyms or x.address in concept.pseudonyms]
                # if equi_targets:
                #     for equi in equi_targets:
                #         concept_map[equi.address] = set()
                concept_map[concept.address].add(input)

    # second loop, include partials
    for input in inputs:
        for concept in input.output_concepts:
            if concept.address not in inherited and not (
                concept.pseudonyms and any(s in inherited for s in concept.pseudonyms)
            ):
                continue
            if (
                isinstance(input, QueryDatasource)
                and concept.address in input.hidden_concepts
            ):
                continue
            if len(concept_map.get(concept.address, [])) == 0:
                concept_map[concept.address].add(input)
    # this adds our new derived metrics, which are not created in this CTE
    for target in targets:
        if target.address not in inherited:
            # an empty source means it is defined in this CTE
            concept_map[target.address] = set()
    return concept_map


def get_all_parent_partial(
    all_concepts: List[BuildConcept], parents: List["StrategyNode"]
) -> List[BuildConcept]:
    return unique(
        [
            c
            for c in all_concepts
            if len(
                [
                    p
                    for p in parents
                    if c.address in [x.address for x in p.partial_concepts]
                ]
            )
            >= 1
            and all(
                [
                    c.address in p.partial_lcl
                    for p in parents
                    if c.address in p.output_lcl
                ]
            )
        ],
        "address",
    )


def get_all_parent_nullable(
    all_concepts: List[BuildConcept], parents: List["StrategyNode"]
) -> List[BuildConcept]:
    return unique(
        [
            c
            for c in all_concepts
            if len(
                [
                    p
                    for p in parents
                    if c.address in [x.address for x in p.nullable_concepts]
                ]
            )
            >= 1
        ],
        "address",
    )


class StrategyNode:
    source_type = SourceType.ABSTRACT

    def __init__(
        self,
        input_concepts: List[BuildConcept],
        output_concepts: List[BuildConcept],
        environment: BuildEnvironment,
        whole_grain: bool = False,
        parents: List["StrategyNode"] | None = None,
        partial_concepts: List[BuildConcept] | None = None,
        nullable_concepts: List[BuildConcept] | None = None,
        depth: int = 0,
        conditions: (
            BuildConditional | BuildComparison | BuildParenthetical | None
        ) = None,
        preexisting_conditions: (
            BuildConditional | BuildComparison | BuildParenthetical | None
        ) = None,
        force_group: bool | None = None,
        grain: Optional[BuildGrain] = None,
        hidden_concepts: set[str] | None = None,
        existence_concepts: List[BuildConcept] | None = None,
        virtual_output_concepts: List[BuildConcept] | None = None,
        ordering: BuildOrderBy | None = None,
    ):
        self.input_concepts: List[BuildConcept] = (
            unique(input_concepts, "address") if input_concepts else []
        )
        self.input_lcl = LooseBuildConceptList(concepts=self.input_concepts)
        self.output_concepts: List[BuildConcept] = unique(output_concepts, "address")
        self.output_lcl = LooseBuildConceptList(concepts=self.output_concepts)

        self.environment = environment
        self.whole_grain = whole_grain
        self.parents = parents or []
        self.resolution_cache: Optional[QueryDatasource] = None

        self.nullable_concepts = nullable_concepts or get_all_parent_nullable(
            self.output_concepts, self.parents
        )
        self.ordering = ordering
        self.depth = depth
        self.conditions = conditions
        self.grain = grain
        self.force_group = force_group
        self.tainted = False
        self.hidden_concepts = hidden_concepts or set()
        self.existence_concepts = existence_concepts or []
        self.virtual_output_concepts = virtual_output_concepts or []
        self.preexisting_conditions = preexisting_conditions
        if self.conditions and not self.preexisting_conditions:
            self.preexisting_conditions = self.conditions
        elif (
            self.conditions
            and self.preexisting_conditions
            and self.conditions != self.preexisting_conditions
        ):
            self.preexisting_conditions = BuildConditional(
                left=self.conditions,
                right=self.preexisting_conditions,
                operator=BooleanOperator.AND,
            )
        self.partial_concepts: list[BuildConcept] = self.derive_partials(
            partial_concepts
        )
        self.validate_inputs()
        self.log = True

    def validate_inputs(self):
        if not self.parents:
            return
        non_hidden = set()
        hidden = set()
        usable_outputs = set()
        for x in self.parents:
            for z in x.usable_outputs:
                usable_outputs.add(z.address)
                non_hidden.add(z.address)
                for psd in z.pseudonyms:
                    non_hidden.add(psd)
            for z in x.hidden_concepts:
                hidden.add(z)
        if not all([x.address in non_hidden for x in self.input_concepts]):
            missing = [x for x in self.input_concepts if x.address not in non_hidden]
            raise ValueError(
                f"Invalid input concepts; {missing} are missing non-hidden parent nodes; have {non_hidden} and hidden {hidden} from root {usable_outputs}"
            )

    def add_parents(self, parents: list["StrategyNode"]):
        self.parents += parents
        self.partial_concepts = self.derive_partials(None)
        return self

    def set_preexisting_conditions(
        self, conditions: BuildConditional | BuildComparison | BuildParenthetical
    ):
        self.preexisting_conditions = conditions
        return self

    def add_condition(
        self, condition: BuildConditional | BuildComparison | BuildParenthetical
    ):
        if self.conditions and condition == self.conditions:
            return self
        if self.conditions:
            self.conditions = BuildConditional(
                left=self.conditions, right=condition, operator=BooleanOperator.AND
            )
        else:
            self.conditions = condition
        self.set_preexisting_conditions(condition)
        self.rebuild_cache()
        return self

    def derive_partials(
        self, partial_concepts: List[BuildConcept] | None = None
    ) -> List[BuildConcept]:
        # validate parents exist
        # assign partial values where needed
        for parent in self.parents:
            if not parent:
                raise SyntaxError("Unresolvable parent")

        # TODO: make this accurate
        if self.parents and partial_concepts is None:
            partials = get_all_parent_partial(self.output_concepts, self.parents)
        elif partial_concepts is None:
            partials = []
        else:
            partials = partial_concepts
        self.partial_lcl = LooseBuildConceptList(concepts=partials)
        return partials

    def add_output_concepts(self, concepts: List[BuildConcept], rebuild: bool = True):
        for concept in concepts:
            if concept.address not in self.output_lcl.addresses:
                self.output_concepts.append(concept)
        self.output_lcl = LooseBuildConceptList(concepts=self.output_concepts)
        if rebuild:
            self.rebuild_cache()
        return self

    def add_partial_concepts(self, concepts: List[BuildConcept], rebuild: bool = True):
        for concept in concepts:
            if concept.address not in self.partial_lcl.addresses:
                self.partial_concepts.append(concept)
        self.partial_lcl = LooseBuildConceptList(concepts=self.partial_concepts)
        if rebuild:
            self.rebuild_cache()
        return self

    def add_existence_concepts(
        self, concepts: List[BuildConcept], rebuild: bool = True
    ):
        for concept in concepts:
            if concept.address not in self.output_concepts:
                self.existence_concepts.append(concept)
        if rebuild:
            self.rebuild_cache()
        return self

    def set_visible_concepts(self, concepts: List[BuildConcept]):
        for x in self.output_concepts:
            if x.address not in [c.address for c in concepts]:
                self.hidden_concepts.add(x.address)
        return self

    def set_output_concepts(
        self,
        concepts: List[BuildConcept],
        rebuild: bool = True,
        change_visibility: bool = True,
    ):
        # exit if no changes
        if self.output_concepts == concepts:
            return self
        self.output_concepts = concepts
        if self.hidden_concepts and change_visibility:
            self.hidden_concepts = set(
                x for x in self.hidden_concepts if x not in concepts
            )

        self.output_lcl = LooseBuildConceptList(concepts=self.output_concepts)

        if rebuild:
            self.rebuild_cache()
        return self

    def add_output_concept(self, concept: BuildConcept, rebuild: bool = True):
        return self.add_output_concepts([concept], rebuild)

    def hide_output_concepts(
        self, concepts: List[BuildConcept] | list[str] | set[str], rebuild: bool = True
    ):
        for x in concepts:
            if isinstance(x, BuildConcept):
                self.hidden_concepts.add(x.address)
            else:
                self.hidden_concepts.add(x)
        if rebuild:
            self.rebuild_cache()
        return self

    def unhide_output_concepts(
        self, concepts: List[BuildConcept], rebuild: bool = True
    ):
        self.hidden_concepts = set(x for x in self.hidden_concepts if x not in concepts)
        if rebuild:
            self.rebuild_cache()
        return self

    @property
    def usable_outputs(self) -> list[BuildConcept]:
        return [
            x for x in self.output_concepts if x.address not in self.hidden_concepts
        ]

    @property
    def logging_prefix(self) -> str:
        return "\t" * self.depth

    @property
    def all_concepts(self) -> list[BuildConcept]:
        return [*self.output_concepts]

    @property
    def all_used_concepts(self) -> list[BuildConcept]:
        return [*self.input_concepts, *self.existence_concepts]

    def __repr__(self):
        concepts = self.all_concepts
        addresses = [c.address for c in concepts]
        contents = ",".join(sorted(addresses[:3]))
        if len(addresses) > 3:
            extra = len(addresses) - 3
            contents += f"...{extra} more"
        return f"{self.__class__.__name__}<{contents}>"

    def _resolve(self) -> QueryDatasource:
        parent_sources: List[QueryDatasource | BuildDatasource] = [
            p.resolve() for p in self.parents
        ]

        grain = (
            self.grain if self.grain else BuildGrain.from_concepts(self.output_concepts)
        )
        source_map = resolve_concept_map(
            parent_sources,
            targets=self.output_concepts,
            inherited_inputs=self.input_concepts + self.existence_concepts,
        )

        return QueryDatasource(
            input_concepts=self.input_concepts,
            output_concepts=self.output_concepts,
            datasources=parent_sources,
            source_type=self.source_type,
            source_map=source_map,
            joins=[],
            grain=grain,
            condition=self.conditions,
            partial_concepts=self.partial_concepts,
            nullable_concepts=self.nullable_concepts,
            force_group=self.force_group,
            hidden_concepts=self.hidden_concepts,
            ordering=self.ordering,
        )

    def rebuild_cache(self) -> QueryDatasource:
        self.tainted = True
        self.output_lcl = LooseBuildConceptList(concepts=self.output_concepts)
        if not self.resolution_cache:
            return self.resolve()
        self.resolution_cache = None
        return self.resolve()

    def resolve(self) -> QueryDatasource:
        if self.resolution_cache:
            return self.resolution_cache
        qds = self._resolve()
        self.resolution_cache = qds
        return qds

    def copy(self) -> "StrategyNode":
        return self.__class__(
            input_concepts=list(self.input_concepts),
            output_concepts=list(self.output_concepts),
            environment=self.environment,
            whole_grain=self.whole_grain,
            parents=list(self.parents),
            partial_concepts=list(self.partial_concepts),
            nullable_concepts=list(self.nullable_concepts),
            depth=self.depth,
            conditions=self.conditions,
            preexisting_conditions=self.preexisting_conditions,
            force_group=self.force_group,
            grain=self.grain,
            hidden_concepts=set(self.hidden_concepts),
            existence_concepts=list(self.existence_concepts),
            virtual_output_concepts=list(self.virtual_output_concepts),
            ordering=self.ordering,
        )


@dataclass
class NodeJoin:
    left_node: StrategyNode
    right_node: StrategyNode
    concepts: List[BuildConcept]
    join_type: JoinType
    filter_to_mutual: bool = False
    concept_pairs: list[ConceptPair] | None = None

    def __post_init__(self):
        if self.left_node == self.right_node:
            raise SyntaxError("Invalid join, left and right nodes are the same")
        if self.concept_pairs:
            return
        final_concepts = []
        for concept in self.concepts:
            include = True
            for ds in [self.left_node, self.right_node]:
                if concept.address not in [c.address for c in ds.all_concepts]:
                    if self.filter_to_mutual:
                        include = False
                    else:
                        raise SyntaxError(
                            f"Invalid join, missing {concept} on {str(ds)}, have"
                            f" {[c.address for c in ds.all_concepts]}"
                        )
            if include:
                final_concepts.append(concept)
        if not final_concepts and self.concepts:
            # if one datasource only has constants
            # we can join on 1=1
            for ds in [self.left_node, self.right_node]:
                if all([c.derivation == Derivation.CONSTANT for c in ds.all_concepts]):
                    self.concepts = []
                    return

            left_keys = [c.address for c in self.left_node.all_concepts]
            right_keys = [c.address for c in self.right_node.all_concepts]
            match_concepts = [c.address for c in self.concepts]
            raise SyntaxError(
                "No mutual join keys found between"
                f" {self.left_node} and"
                f" {self.right_node}, left_keys {left_keys},"
                f" right_keys {right_keys},"
                f" provided join concepts {match_concepts}"
            )
        self.concepts = final_concepts

    @property
    def unique_id(self) -> str:
        nodes = sorted([self.left_node, self.right_node], key=lambda x: str(x))
        return str(nodes) + self.join_type.value

    def __str__(self):
        return (
            f"{self.join_type.value} JOIN {self.left_node} and"
            f" {self.right_node} on"
            f" {','.join([str(k) for k in self.concepts])}"
        )


class WhereSafetyNode(StrategyNode):
    """Specialized node to be used to pad certain
    select outputs that can't be immediately used in a where
    clause; eg window functions. Will remove itself if not required."""

    def resolve(self) -> QueryDatasource:
        if not self.conditions and len(self.parents) == 1:
            parent = self.parents[0]
            parent = parent.copy()
            # avoid performance hit by not rebuilding until end
            parent.set_output_concepts(self.output_concepts, rebuild=False)

            # these conditions
            if self.preexisting_conditions:
                parent.set_preexisting_conditions(self.preexisting_conditions)
            # TODO: add a helper for this
            parent.ordering = self.ordering

            # actually build the node
            parent.rebuild_cache()
            return parent.resolve()
        return super().resolve()



from .basic_node import gen_basic_node
from .constant_node import gen_constant_node
from .filter_node import gen_filter_node
from .group_node import gen_group_node
from .group_to_node import gen_group_to_node
from .multiselect_node import gen_multiselect_node
from .node_merge_node import gen_merge_node
from .recursive_node import gen_recursive_node
from .rowset_node import gen_rowset_node
from .select_node import gen_select_node
from .synonym_node import gen_synonym_node
from .union_node import gen_union_node
from .unnest_node import gen_unnest_node
from .window_node import gen_window_node

__all__ = [
    "gen_filter_node",
    "gen_window_node",
    "gen_group_node",
    "gen_select_node",
    "gen_basic_node",
    "gen_unnest_node",
    "gen_union_node",
    "gen_merge_node",
    "gen_group_to_node",
    "gen_rowset_node",
    "gen_multiselect_node",
    "gen_synonym_node",
    "gen_recursive_node",
    "gen_constant_node",
]



from trilogy.core.models.execute import (
    CTE,
    QueryDatasource,
    UnionCTE,
)
from trilogy.core.processing.nodes import StrategyNode
from trilogy.core.statements.author import (
    MultiSelectStatement,
    PersistStatement,
    RowsetDerivationStatement,
    SelectStatement,
)


class BaseHook:
    pass

    def process_multiselect_info(self, select: MultiSelectStatement):
        print("Multiselect with components:")
        for x in select.selects:
            self.process_select_info(x)

    def process_select_info(self, select: SelectStatement):
        print(f"Select statement grain: {str(select.grain)}")

    def process_persist_info(self, persist: PersistStatement):
        print(f"Persist statement persisting to {persist.address}")
        self.process_select_info(persist.select)

    def process_rowset_info(self, rowset: RowsetDerivationStatement):
        print(f"Rowset statement with grain {str(rowset.select.grain)}")

    def process_root_datasource(self, datasource: QueryDatasource):
        pass

    def process_root_cte(self, cte: CTE | UnionCTE):
        pass

    def process_root_strategy_node(self, node: StrategyNode):
        pass



from trilogy.core.constants import ALL_ROWS_CONCEPT, INTERNAL_NAMESPACE
from trilogy.core.enums import Derivation, FunctionType, Granularity, Purpose
from trilogy.core.models.author import Concept, Function, Grain
from trilogy.core.models.core import DataType

DEFAULT_CONCEPTS = {
    ALL_ROWS_CONCEPT: Concept(
        name=ALL_ROWS_CONCEPT,
        namespace=INTERNAL_NAMESPACE,
        datatype=DataType.INTEGER,
        purpose=Purpose.CONSTANT,
        grain=Grain(),
        lineage=Function(
            operator=FunctionType.CONSTANT,
            arguments=[1],
            output_datatype=DataType.INTEGER,
            output_purpose=Purpose.CONSTANT,
        ),
        granularity=Granularity.SINGLE_ROW,
        derivation=Derivation.CONSTANT,
    ),
    "concept_address": Concept(
        name="concept_address",
        namespace=INTERNAL_NAMESPACE,
        datatype=DataType.STRING,
        purpose=Purpose.KEY,
        grain=Grain(),
        granularity=Granularity.SINGLE_ROW,
        derivation=Derivation.CONSTANT,
    ),
    "concept_datatype": Concept(
        name="concept_datatype",
        namespace=INTERNAL_NAMESPACE,
        datatype=DataType.STRING,
        purpose=Purpose.KEY,
        grain=Grain(),
        granularity=Granularity.SINGLE_ROW,
        derivation=Derivation.CONSTANT,
    ),
    "concept_description": Concept(
        name="concept_description",
        namespace=INTERNAL_NAMESPACE,
        datatype=DataType.STRING,
        purpose=Purpose.KEY,
        grain=Grain(),
        granularity=Granularity.SINGLE_ROW,
        derivation=Derivation.CONSTANT,
    ),
    "datasource": Concept(
        name="datasource",
        namespace=INTERNAL_NAMESPACE,
        datatype=DataType.STRING,
        purpose=Purpose.KEY,
        grain=Grain(),
        granularity=Granularity.SINGLE_ROW,
        derivation=Derivation.CONSTANT,
    ),
    "query_text": Concept(
        name="query_text",
        namespace=INTERNAL_NAMESPACE,
        datatype=DataType.STRING,
        purpose=Purpose.KEY,
        grain=Grain(),
        granularity=Granularity.SINGLE_ROW,
        derivation=Derivation.CONSTANT,
    ),
    "label": Concept(
        name="label",
        namespace=INTERNAL_NAMESPACE,
        datatype=DataType.STRING,
        purpose=Purpose.KEY,
        grain=Grain(),
        granularity=Granularity.SINGLE_ROW,
        derivation=Derivation.CONSTANT,
    ),
    "expected": Concept(
        name="expected_value",
        namespace=INTERNAL_NAMESPACE,
        datatype=DataType.STRING,
        purpose=Purpose.KEY,
        grain=Grain(),
        granularity=Granularity.SINGLE_ROW,
        derivation=Derivation.CONSTANT,
    ),
}



from __future__ import annotations

from abc import ABC
from collections import UserDict, UserList
from datetime import date, datetime
from decimal import Decimal
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    get_args,
)

from pydantic import (
    BaseModel,
    ConfigDict,
    field_validator,
)
from pydantic_core import core_schema

from trilogy.constants import (
    MagicConstants,
)
from trilogy.core.enums import Modifier, Ordering


class DataTyped(ABC):

    # this is not abstract
    # only because when it's a pydantic property, it fails validation
    @property
    def output_datatype(self) -> CONCRETE_TYPES:  # type: ignore
        """
        This is a huge hack to get property vs pydantic attribute inheritance to work.
        """
        if "output_datatype" in self.__dict__:
            return self.__dict__["output_datatype"]
        raise NotImplementedError


class Addressable(ABC):

    @property
    def _address(self):
        return self.address


TYPEDEF_TYPES = Union[
    "DataType",
    "MapType",
    "ArrayType",
    "NumericType",
    "StructType",
    "DataTyped",
    "TraitDataType",
]

CONCRETE_TYPES = Union[
    "DataType",
    "MapType",
    "ArrayType",
    "NumericType",
    "StructType",
    "TraitDataType",
]

KT = TypeVar("KT")
VT = TypeVar("VT")
LT = TypeVar("LT")


class DataType(Enum):
    # PRIMITIVES
    STRING = "string"
    BOOL = "bool"
    MAP = "map"
    NUMBER = "number"
    FLOAT = "float"
    NUMERIC = "numeric"
    INTEGER = "int"
    BIGINT = "bigint"
    DATE = "date"
    DATETIME = "datetime"
    TIMESTAMP = "timestamp"
    ARRAY = "array"
    DATE_PART = "date_part"
    STRUCT = "struct"
    NULL = "null"

    # GRANULAR
    UNIX_SECONDS = "unix_seconds"

    # PARSING
    UNKNOWN = "unknown"

    @property
    def data_type(self):
        return self

    def __str__(self) -> str:
        return self.name


class TraitDataType(BaseModel):
    type: DataType | NumericType | StructType | ArrayType | MapType
    traits: list[str]

    def __hash__(self):
        return hash(self.type)

    def __str__(self) -> str:
        return f"Trait<{self.type}, {self.traits}>"

    def __eq__(self, other):
        if isinstance(other, DataType):
            return self.type == other
        elif isinstance(other, TraitDataType):
            return self.type == other.type and self.traits == other.traits
        return False

    @property
    def data_type(self):
        return self.type

    @property
    def value(self):
        return self.data_type.value


class NumericType(BaseModel):
    precision: int = 20
    scale: int = 5

    def __str__(self) -> str:
        return f"Numeric({self.precision},{self.scale})"

    @property
    def data_type(self):
        return DataType.NUMERIC

    @property
    def value(self):
        return self.data_type.value


class ArrayType(BaseModel):
    model_config = ConfigDict(frozen=True)
    type: TYPEDEF_TYPES

    @field_validator("type", mode="plain")
    def validate_type(cls, v):
        return v

    def __str__(self) -> str:
        return f"ListType<{self.type}>"

    @property
    def data_type(self):
        return DataType.ARRAY

    @property
    def value(self):
        return self.data_type.value

    @property
    def value_data_type(
        self,
    ) -> CONCRETE_TYPES:
        if isinstance(self.type, DataTyped):
            return self.type.output_datatype
        return self.type


class MapType(BaseModel):
    key_type: TYPEDEF_TYPES
    value_type: TYPEDEF_TYPES

    @field_validator("value_type", mode="plain")
    def validate_type(cls, v):
        return v

    @field_validator("key_type", mode="plain")
    def validate_key_type(cls, v):
        return v

    @property
    def data_type(self):
        return DataType.MAP

    @property
    def value(self):
        return self.data_type.value

    @property
    def value_data_type(
        self,
    ) -> CONCRETE_TYPES:
        if isinstance(self.value_type, DataTyped):
            return self.value_type.output_datatype
        return self.value_type

    @property
    def key_data_type(
        self,
    ) -> CONCRETE_TYPES:
        if isinstance(self.key_type, DataTyped):
            return self.key_type.output_datatype
        return self.key_type


class StructComponent(BaseModel):
    name: str
    type: TYPEDEF_TYPES
    modifiers: list[Modifier] = []

    @field_validator("type", mode="plain")
    def validate_Type(cls, v):
        return v


class StructType(BaseModel):
    fields: Sequence[StructComponent | TYPEDEF_TYPES]
    fields_map: Dict[str, DataTyped | int | float | str | StructComponent]

    def __repr__(self):
        return "struct<{}>".format(
            ", ".join(
                f"{field.name}:{field.type.name}"
                for field in self.fields
                if isinstance(field, StructComponent)
            )
        )

    def __str__(self) -> str:
        return self.__repr__()

    @field_validator("fields", mode="plain")
    def validate_type(cls, v):
        final = []
        for field in v:
            final.append(field)
        return final

    @field_validator("fields_map", mode="plain")
    def validate_fields_map(cls, v):
        return v

    @property
    def data_type(self):
        return DataType.STRUCT

    @property
    def value(self):
        return self.data_type.value

    @property
    def field_types(self) -> Dict[str, CONCRETE_TYPES]:
        out: Dict[str, CONCRETE_TYPES] = {}
        keys = list(self.fields_map.keys())
        for idx, field in enumerate(self.fields):
            if isinstance(field, StructComponent):
                out[field.name] = arg_to_datatype(field.type)
            elif isinstance(field, DataTyped):
                out[keys[idx]] = field.output_datatype
            else:
                out[keys[idx]] = field
        return out

    def __hash__(self):
        return hash(str(self))


class ListWrapper(Generic[VT], UserList):
    """Used to distinguish parsed list objects from other lists"""

    def __init__(self, *args, type: DataType, nullable: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = type
        self.nullable = nullable

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: Callable[[Any], core_schema.CoreSchema]
    ) -> core_schema.CoreSchema:
        args = get_args(source_type)
        if args:
            schema = handler(List[args])  # type: ignore
        else:
            schema = handler(List)
        return core_schema.no_info_after_validator_function(cls.validate, schema)

    @classmethod
    def validate(cls, v):
        return cls(v, type=arg_to_datatype(v[0]))


class MapWrapper(Generic[KT, VT], UserDict):
    """Used to distinguish parsed map objects from other dicts"""

    def __init__(self, *args, key_type: DataType, value_type: DataType, **kwargs):
        super().__init__(*args, **kwargs)
        self.key_type = key_type
        self.value_type = value_type

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: Callable[[Any], core_schema.CoreSchema]
    ) -> core_schema.CoreSchema:
        args = get_args(source_type)
        if args:
            schema = handler(Dict[args])  # type: ignore
        else:
            schema = handler(Dict)
        return core_schema.no_info_after_validator_function(cls.validate, schema)

    @classmethod
    def validate(cls, v):
        return cls(
            v,
            key_type=arg_to_datatype(list(v.keys()).pop()),
            value_type=arg_to_datatype(list(v.values()).pop()),
        )


class TupleWrapper(Generic[VT], tuple):
    """Used to distinguish parsed tuple objects from other tuples"""

    def __init__(self, val, type: DataType, nullable: bool = False, **kwargs):
        super().__init__()
        self.type = type
        self.val = val
        self.nullable = nullable

    def __getnewargs__(self):
        return (self.val, self.type)

    def __new__(cls, val, type: DataType, **kwargs):
        return super().__new__(cls, tuple(val))
        # self.type = type

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: Callable[[Any], core_schema.CoreSchema]
    ) -> core_schema.CoreSchema:
        args = get_args(source_type)
        if args:
            schema = handler(Tuple[args])  # type: ignore
        else:
            schema = handler(Tuple)
        return core_schema.no_info_after_validator_function(cls.validate, schema)

    @classmethod
    def validate(cls, v):
        return cls(v, type=arg_to_datatype(v[0]))


def list_to_wrapper(args):
    rtypes = [arg_to_datatype(arg) for arg in args]
    types = [arg for arg in rtypes if arg != DataType.NULL]
    if not len(set(types)) == 1:
        raise SyntaxError(f"Cannot create a list with this set of types: {set(types)}")
    return ListWrapper(args, type=types[0], nullable=DataType.NULL in rtypes)


def tuple_to_wrapper(args):
    rtypes = [arg_to_datatype(arg) for arg in args]
    types = [arg for arg in rtypes if arg != DataType.NULL]
    if not len(set(types)) == 1:
        raise SyntaxError(f"Cannot create a tuple with this set of types: {set(types)}")
    return TupleWrapper(args, type=types[0], nullable=DataType.NULL in rtypes)


def dict_to_map_wrapper(arg):
    key_types = [arg_to_datatype(arg) for arg in arg.keys()]

    value_types = [arg_to_datatype(arg) for arg in arg.values()]
    assert len(set(key_types)) == 1
    assert len(set(key_types)) == 1
    return MapWrapper(arg, key_type=key_types[0], value_type=value_types[0])


def merge_datatypes(
    inputs: list[
        DataType | ArrayType | StructType | MapType | NumericType | TraitDataType
    ],
) -> DataType | ArrayType | StructType | MapType | NumericType | TraitDataType:
    """This is a temporary hack for doing between
    allowable datatype transformation matrix"""
    if len(inputs) == 1:
        return inputs[0]
    if set(inputs) == {DataType.INTEGER, DataType.FLOAT}:
        return DataType.FLOAT
    if set(inputs) == {DataType.INTEGER, DataType.NUMERIC}:
        return DataType.NUMERIC
    if any(isinstance(x, NumericType) for x in inputs) and all(
        isinstance(x, NumericType)
        or x in (DataType.INTEGER, DataType.FLOAT, DataType.NUMERIC)
        for x in inputs
    ):
        candidate = next(x for x in inputs if isinstance(x, NumericType))
        return candidate
    return inputs[0]


def is_compatible_datatype(left, right):
    # for unknown types, we can't make any assumptions
    if all(
        isinstance(x, NumericType)
        or x in (DataType.INTEGER, DataType.FLOAT, DataType.NUMERIC)
        for x in (left, right)
    ):
        return True
    elif isinstance(left, NumericType) or isinstance(right, NumericType):
        return False
    if right == DataType.UNKNOWN or left == DataType.UNKNOWN:
        return True
    if left == right:
        return True
    if {left, right} == {DataType.NUMERIC, DataType.FLOAT}:
        return True

    if {left, right} == {DataType.NUMERIC, DataType.INTEGER}:
        return True
    if {left, right} == {DataType.FLOAT, DataType.INTEGER}:
        return True
    return False


def arg_to_datatype(arg) -> CONCRETE_TYPES:

    if isinstance(arg, MagicConstants):
        if arg == MagicConstants.NULL:
            return DataType.NULL
        raise ValueError(f"Cannot parse arg datatype for arg of type {arg}")
    elif isinstance(arg, bool):
        return DataType.BOOL
    elif isinstance(arg, Ordering):
        return DataType.STRING  # TODO: revisit
    elif isinstance(arg, int):
        return DataType.INTEGER
    elif isinstance(arg, str):
        return DataType.STRING
    elif isinstance(arg, float):
        return DataType.FLOAT
    elif isinstance(arg, Decimal):
        return DataType.NUMERIC
    elif isinstance(arg, DataType):
        return arg
    elif isinstance(arg, NumericType):
        return arg
    elif isinstance(arg, TraitDataType):
        return arg
    elif isinstance(arg, ListWrapper):
        return ArrayType(type=arg.type)
    elif isinstance(arg, DataTyped):
        return arg.output_datatype
    elif isinstance(arg, TupleWrapper):
        return ArrayType(type=arg.type)
    elif isinstance(arg, list):
        wrapper = list_to_wrapper(arg)
        return ArrayType(type=wrapper.type)
    elif isinstance(arg, MapWrapper):
        return MapType(key_type=arg.key_type, value_type=arg.value_type)
    elif isinstance(arg, datetime):
        return DataType.DATETIME
    elif isinstance(arg, date):
        return DataType.DATE
    elif isinstance(arg, StructComponent):
        return arg_to_datatype(arg.type)
    else:
        raise ValueError(
            f"Cannot parse arg datatype for arg of raw type {type(arg)} value {arg}"
        )



from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    try:
        from pandas import DataFrame
    except ImportError:
        DataFrame = Any


class DialectConfig:
    def __init__(self):
        pass

    def connection_string(self) -> str:
        raise NotImplementedError

    @property
    def connect_args(self) -> dict:
        return {}


class BigQueryConfig(DialectConfig):
    def __init__(self, project: str, client):
        self.project = project
        self.client = client

    def connection_string(self) -> str:
        return f"bigquery://{self.project}?user_supplied_client=True"

    @property
    def connect_args(self) -> dict:
        return {"client": self.client}


class DuckDBConfig(DialectConfig):
    def __init__(self, path: str | None = None):
        self.path = path

    def connection_string(self) -> str:
        if not self.path:
            return "duckdb:///:memory:"
        return f"duckdb:///{self.path}"


class PostgresConfig(DialectConfig):
    def __init__(
        self, host: str, port: int, username: str, password: str, database: str
    ):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.database = database

    def connection_string(self) -> str:
        return f"postgresql://{self.username}:{self.password}@{self.host}:{self.port}"


class SQLServerConfig(DialectConfig):
    def __init__(
        self, host: str, port: int, username: str, password: str, database: str
    ):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.database = database

    def connection_string(self) -> str:
        return f"sqlserver//{self.username}:{self.password}@{self.host}:{self.port}"


class SnowflakeConfig(DialectConfig):
    def __init__(
        self,
        account: str,
        username: str,
        password: str,
        database: str | None = None,
        schema: str | None = None,
    ):
        self.account = account
        self.username = username
        self.password = password
        self.database = database
        self.schema = schema
        if self.schema and not self.database:
            raise ValueError("Setting snowflake schema also requires setting database")

    def connection_string(self) -> str:
        if self.schema:
            return f"snowflake://{self.username}:{self.password}@{self.account}/{self.database}/{self.schema}"
        if self.database:
            return f"snowflake://{self.username}:{self.password}@{self.account}/{self.database}"
        return f"snowflake://{self.username}:{self.password}@{self.account}"


class PrestoConfig(DialectConfig):
    def __init__(
        self,
        host: str,
        port: int,
        username: str,
        password: str,
        catalog: str,
        schema: str | None = None,
    ):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.catalog = catalog
        self.schema = schema

    def connection_string(self) -> str:
        if self.schema:
            return f"presto://{self.username}:{self.password}@{self.host}:{self.port}/{self.catalog}/{self.schema}"
        return f"presto://{self.username}:{self.password}@{self.host}:{self.port}/{self.catalog}"


class TrinoConfig(PrestoConfig):
    def connection_string(self) -> str:
        if self.schema:
            return f"trino://{self.username}:{self.password}@{self.host}:{self.port}/{self.catalog}/{self.schema}"
        return f"trino://{self.username}:{self.password}@{self.host}:{self.port}/{self.catalog}"


class DataFrameConfig(DuckDBConfig):
    def __init__(self, dataframes: dict[str, "DataFrame"]):
        super().__init__()
        self.dataframes = dataframes



from typing import List

from trilogy.constants import logger
from trilogy.core.enums import Derivation, Purpose
from trilogy.core.models.build import (
    BuildConcept,
    BuildDatasource,
    BuildFilterItem,
    BuildGrain,
    BuildRowsetItem,
)
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.models.execute import QueryDatasource, UnnestJoin
from trilogy.core.processing.nodes import GroupNode, MergeNode, StrategyNode
from trilogy.core.processing.utility import GroupRequiredResponse


def depth_to_prefix(depth: int) -> str:
    return "\t" * depth


LOGGER_PREFIX = "[DISCOVERY LOOP]"


def calculate_effective_parent_grain(
    node: QueryDatasource | BuildDatasource,
) -> BuildGrain:
    # calculate the effective grain of the parent node
    # this is the union of all parent grains
    if isinstance(node, QueryDatasource):
        grain = BuildGrain()
        qds = node
        if not qds.joins:
            return qds.datasources[0].grain
        seen = set()
        for join in qds.joins:
            if isinstance(join, UnnestJoin):
                grain += BuildGrain(components=set([x.address for x in join.concepts]))
                continue
            pairs = join.concept_pairs or []
            for key in pairs:
                left = key.existing_datasource
                logger.info(f"adding left grain {left.grain} for join key {key.left}")
                grain += left.grain
                seen.add(left.name)
            keys = [key.right for key in pairs]
            join_grain = BuildGrain.from_concepts(keys)
            if join_grain == join.right_datasource.grain:
                logger.info(f"irrelevant right join {join}, does not change grain")
            else:
                logger.info(
                    f"join changes grain, adding {join.right_datasource.grain} to {grain}"
                )
                grain += join.right_datasource.grain
            seen.add(join.right_datasource.name)
        for x in qds.datasources:
            # if we haven't seen it, it's still contributing to grain
            # unless used ONLY in a subselect
            # so the existence check is a [bad] proxy for that
            if x.name not in seen and not (
                qds.condition
                and qds.condition.existence_arguments
                and any(
                    [
                        c.address in block
                        for c in x.output_concepts
                        for block in qds.condition.existence_arguments
                    ]
                )
            ):
                logger.info(f"adding unjoined grain {x.grain} for datasource {x.name}")
                grain += x.grain
        return grain
    else:
        return node.grain or BuildGrain()


def check_if_group_required(
    downstream_concepts: List[BuildConcept],
    parents: list[QueryDatasource | BuildDatasource],
    environment: BuildEnvironment,
    depth: int = 0,
) -> GroupRequiredResponse:
    padding = "\t" * depth
    target_grain = BuildGrain.from_concepts(
        downstream_concepts,
        environment=environment,
    )

    comp_grain = BuildGrain()
    for source in parents:
        # comp_grain += source.grain
        comp_grain += calculate_effective_parent_grain(source)

    # dynamically select if we need to group
    # we must avoid grouping if we are already at grain
    if comp_grain.issubset(target_grain):

        logger.info(
            f"{padding}{LOGGER_PREFIX} Group requirement check:  {comp_grain}, target: {target_grain}, grain is subset of target, no group node required"
        )
        return GroupRequiredResponse(target_grain, comp_grain, False)
    # find out what extra is in the comp grain vs target grain
    difference = [
        environment.concepts[c] for c in (comp_grain - target_grain).components
    ]
    logger.info(
        f"{padding}{LOGGER_PREFIX} Group requirement check: upstream grain: {comp_grain}, desired grain: {target_grain} from , difference {[x.address for x in difference]}"
    )
    for x in difference:
        logger.info(
            f"{padding}{LOGGER_PREFIX} Difference concept {x.address} purpose {x.purpose} keys {x.keys}"
        )

    # if the difference is all unique properties whose keys are in the source grain
    # we can also suppress the group
    if all(
        [
            x.keys
            and all(
                environment.concepts[z].address in comp_grain.components for z in x.keys
            )
            for x in difference
        ]
    ):
        logger.info(
            f"{padding}{LOGGER_PREFIX} Group requirement check: skipped due to unique property validation"
        )
        return GroupRequiredResponse(target_grain, comp_grain, False)
    if all([x.purpose == Purpose.KEY for x in difference]):
        logger.info(
            f"{padding}{LOGGER_PREFIX} checking if downstream is unique properties of key"
        )
        replaced_grain_raw: list[set[str]] = [
            (
                x.keys or set()
                if x.purpose == Purpose.UNIQUE_PROPERTY
                else set([x.address])
            )
            for x in downstream_concepts
            if x.address in target_grain.components
        ]
        # flatten the list of lists
        replaced_grain = [item for sublist in replaced_grain_raw for item in sublist]
        # if the replaced grain is a subset of the comp grain, we can skip the group
        unique_grain_comp = BuildGrain.from_concepts(
            replaced_grain, environment=environment
        )
        if comp_grain.issubset(unique_grain_comp):
            logger.info(
                f"{padding}{LOGGER_PREFIX} Group requirement check: skipped due to unique property validation"
            )
            return GroupRequiredResponse(target_grain, comp_grain, False)
    logger.info(
        f"{padding}{LOGGER_PREFIX} Checking for grain equivalence for filters and rowsets"
    )
    ngrain = []
    for con in target_grain.components:
        full = environment.concepts[con]
        if full.derivation == Derivation.ROWSET:
            ngrain.append(full.address.split(".", 1)[1])
        elif full.derivation == Derivation.FILTER:
            assert isinstance(full.lineage, BuildFilterItem)
            if isinstance(full.lineage.content, BuildConcept):
                ngrain.append(full.lineage.content.address)
        else:
            ngrain.append(full.address)
    target_grain2 = BuildGrain.from_concepts(
        ngrain,
        environment=environment,
    )
    if comp_grain.issubset(target_grain2):
        logger.info(
            f"{padding}{LOGGER_PREFIX} Group requirement check: {comp_grain}, {target_grain2}, pre rowset grain is subset of target, no group node required"
        )
        return GroupRequiredResponse(target_grain2, comp_grain, False)

    logger.info(f"{padding}{LOGGER_PREFIX} Group requirement check: group required")
    return GroupRequiredResponse(target_grain, comp_grain, True)


def group_if_required_v2(
    root: StrategyNode, final: List[BuildConcept], environment: BuildEnvironment
):
    required = check_if_group_required(
        downstream_concepts=final, parents=[root.resolve()], environment=environment
    )
    targets = [
        x
        for x in root.output_concepts
        if x.address in final or any(c in final for c in x.pseudonyms)
    ]
    if required.required:
        if isinstance(root, MergeNode):
            root.force_group = True
            root.set_output_concepts(targets, rebuild=False, change_visibility=False)
            root.rebuild_cache()
            return root
        elif isinstance(root, GroupNode):
            # root.set_output_concepts(final, rebuild=False)
            # root.rebuild_cache()
            return root
        return GroupNode(
            output_concepts=targets,
            input_concepts=targets,
            environment=environment,
            parents=[root],
            partial_concepts=root.partial_concepts,
            preexisting_conditions=root.preexisting_conditions,
        )
    elif isinstance(root, GroupNode):

        return root
    else:
        root.set_output_concepts(targets, rebuild=False, change_visibility=False)
    return root


def get_upstream_concepts(base: BuildConcept, nested: bool = False) -> set[str]:
    upstream = set()
    if nested:
        upstream.add(base.address)
    if not base.lineage:
        return upstream
    for x in base.lineage.concept_arguments:
        # if it's derived from any value in a rowset, ALL rowset items are upstream
        if x.derivation == Derivation.ROWSET:
            assert isinstance(x.lineage, BuildRowsetItem), type(x.lineage)
            for y in x.lineage.rowset.select.output_components:
                upstream.add(f"{x.lineage.rowset.name}.{y.address}")
                # upstream = upstream.union(get_upstream_concepts(y, nested=True))
        upstream = upstream.union(get_upstream_concepts(x, nested=True))
    return upstream


def get_priority_concept(
    all_concepts: List[BuildConcept],
    attempted_addresses: set[str],
    found_concepts: set[str],
    depth: int,
) -> BuildConcept:
    # optimized search for missing concepts
    pass_one = sorted(
        [
            c
            for c in all_concepts
            if c.address not in attempted_addresses and c.address not in found_concepts
        ],
        key=lambda x: x.address,
    )
    # sometimes we need to scan intermediate concepts to get merge keys or filter keys,
    # so do an exhaustive search
    # pass_two = [c for c in all_concepts+filter_only if c.address not in attempted_addresses]
    for remaining_concept in (pass_one,):
        priority = (
            # then multiselects to remove them from scope
            [c for c in remaining_concept if c.derivation == Derivation.MULTISELECT]
            +
            # then rowsets to remove them from scope, as they cannot get partials
            [c for c in remaining_concept if c.derivation == Derivation.ROWSET]
            +
            # then rowsets to remove them from scope, as they cannot get partials
            [c for c in remaining_concept if c.derivation == Derivation.UNION]
            # we should be home-free here
            +
            # then aggregates to remove them from scope, as they cannot get partials
            [c for c in remaining_concept if c.derivation == Derivation.AGGREGATE]
            # then windows to remove them from scope, as they cannot get partials
            + [c for c in remaining_concept if c.derivation == Derivation.WINDOW]
            # then filters to remove them from scope, also cannot get partials
            + [c for c in remaining_concept if c.derivation == Derivation.FILTER]
            # unnests are weird?
            + [c for c in remaining_concept if c.derivation == Derivation.UNNEST]
            + [c for c in remaining_concept if c.derivation == Derivation.RECURSIVE]
            + [c for c in remaining_concept if c.derivation == Derivation.BASIC]
            + [c for c in remaining_concept if c.derivation == Derivation.GROUP_TO]
            + [c for c in remaining_concept if c.derivation == Derivation.CONSTANT]
            # finally our plain selects
            + [
                c for c in remaining_concept if c.derivation == Derivation.ROOT
            ]  # and any non-single row constants
        )

        priority += [
            c
            for c in remaining_concept
            if c.address not in [x.address for x in priority]
        ]
        final = []
        # if any thing is derived from another concept
        # get the derived copy first
        # as this will usually resolve cleaner
        for x in priority:
            if any(
                [
                    x.address
                    in get_upstream_concepts(
                        c,
                    )
                    for c in priority
                ]
            ):
                logger.info(
                    f"{depth_to_prefix(depth)}{LOGGER_PREFIX} delaying fetch of {x.address} as parent of another concept"
                )
                continue
            final.append(x)
        # then append anything we didn't get
        for x2 in priority:
            if x2 not in final:
                final.append(x2)
        if final:
            return final[0]
    raise ValueError(
        f"Cannot resolve query. No remaining priority concepts, have attempted {attempted_addresses}"
    )



from typing import List

from trilogy.constants import logger
from trilogy.core.models.build import (
    BuildConcept,
    BuildFunction,
    BuildWhereClause,
)
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.processing.nodes import (
    History,
    StrategyNode,
    UnnestNode,
    WhereSafetyNode,
)
from trilogy.core.processing.utility import padding

LOGGER_PREFIX = "[GEN_UNNEST_NODE]"


def gen_unnest_node(
    concept: BuildConcept,
    local_optional: List[BuildConcept],
    history: History,
    environment: BuildEnvironment,
    g,
    depth: int,
    source_concepts,
    conditions: BuildWhereClause | None = None,
) -> StrategyNode | None:
    arguments = []
    depth_prefix = "\t" * depth
    if isinstance(concept.lineage, BuildFunction):
        arguments = concept.lineage.concept_arguments

    equivalent_optional = [x for x in local_optional if x.lineage == concept.lineage]

    non_equivalent_optional = [
        x for x in local_optional if x not in equivalent_optional
    ]
    all_parents = arguments + non_equivalent_optional
    logger.info(
        f"{depth_prefix}{LOGGER_PREFIX} unnest node for {concept} with lineage {concept.lineage} has parents + optional {all_parents} and equivalent optional {equivalent_optional}"
    )
    local_conditions = False
    expected_outputs = [concept] + local_optional
    if arguments or local_optional:
        parent = source_concepts(
            mandatory_list=all_parents,
            environment=environment,
            g=g,
            depth=depth + 1,
            history=history,
            conditions=conditions,
        )
        if not parent:
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} could not find unnest node parents"
            )
            return None
    elif conditions:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} unnest node has no parents but conditions inputs {conditions.row_arguments} vs expected output {expected_outputs}"
        )
        if all([x.address in expected_outputs for x in conditions.row_arguments]):
            local_conditions = True
        else:
            parent = source_concepts(
                mandatory_list=conditions.conditional.row_arguments,
                environment=environment,
                g=g,
                depth=depth + 1,
                history=history,
                conditions=conditions,
            )
            if not parent:
                logger.info(
                    f"{padding(depth)}{LOGGER_PREFIX} could not find unnest node condition inputs with no parents"
                )
                return None
    else:
        parent = None
    logger.info(
        f"{depth_prefix}{LOGGER_PREFIX} unnest node for {concept} got parent {parent}"
    )
    base = UnnestNode(
        unnest_concepts=[concept] + equivalent_optional,
        input_concepts=arguments + non_equivalent_optional,
        output_concepts=[concept] + local_optional,
        environment=environment,
        parents=([parent] if parent else []),
    )
    # we need to sometimes nest an unnest node,
    # as unnest operations are not valid in all situations
    # TODO: inline this node when we can detect it's safe
    conditional = conditions.conditional if conditions else None
    new = WhereSafetyNode(
        input_concepts=base.output_concepts,
        output_concepts=base.output_concepts,
        environment=environment,
        parents=[base],
        conditions=conditional if local_conditions is True else None,
        preexisting_conditions=(
            conditional if conditional and local_conditions is False else None
        ),
    )
    # qds = new.resolve()
    # assert qds.source_map[concept.address] == {base.resolve()}
    # for x in equivalent_optional:
    #     assert qds.source_map[x.address] == {base.resolve()}
    return new



from trilogy import Dialects
from trilogy.constants import MagicConstants
from trilogy.core.enums import BooleanOperator, ComparisonOperator, Purpose
from trilogy.core.functions import argument_to_purpose, function_args_to_output_purpose
from trilogy.core.models.author import (
    Comparison,
    Conditional,
    SubselectComparison,
)
from trilogy.core.models.build import BuildComparison
from trilogy.core.models.core import (
    DataType,
    TupleWrapper,
)
from trilogy.core.models.datasource import Datasource
from trilogy.core.models.environment import (
    DictImportResolver,
    Environment,
    EnvironmentOptions,
)
from trilogy.core.statements.author import SelectStatement, ShowStatement
from trilogy.core.statements.execute import ProcessedQuery
from trilogy.dialect.base import BaseDialect
from trilogy.parsing.common import atom_is_relevant
from trilogy.parsing.parse_engine import (
    arg_to_datatype,
    parse_text,
)


def test_in():
    _, parsed = parse_text(
        "const order_id <- 3; SELECT order_id  WHERE order_id IN (1,2,3);"
    )
    query = parsed[-1]
    right = query.where_clause.conditional.right
    assert isinstance(
        right,
        TupleWrapper,
    ), type(right)
    assert right[0] == 1
    rendered = BaseDialect().render_expr(right)
    assert rendered.strip() == "(1,2,3)".strip()

    _, parsed = parse_text(
        "const order_id <- 3; SELECT order_id  WHERE order_id IN (1,);"
    )
    query = parsed[-1]
    right = query.where_clause.conditional.right
    assert isinstance(
        right,
        TupleWrapper,
    ), type(right)
    assert right[0] == 1
    rendered = BaseDialect().render_expr(right)
    assert rendered.strip() == "(1)".strip()


def test_not_in():
    _, parsed = parse_text(
        "const order_id <- 4; SELECT order_id  WHERE order_id NOT IN (1,2,3);"
    )
    query: SelectStatement = parsed[-1]
    right = query.where_clause.conditional.right
    assert isinstance(right, TupleWrapper), type(right)
    assert right[0] == 1
    rendered = BaseDialect().render_expr(right)
    assert rendered.strip() == "(1,2,3)".strip()


def test_datetime_lit_rendering():
    env, parsed = parse_text("const order_id <- 4;")

    from datetime import datetime

    now = datetime.now()
    wrapper = TupleWrapper(
        (now,),
        type=DataType.DATETIME,
    )
    assert wrapper[0] == now, wrapper
    rendered = BaseDialect().render_expr(
        BuildComparison(
            left=env.materialize_for_select().concepts["order_id"],
            operator=ComparisonOperator.NOT_IN,
            right=wrapper,
        )
    )
    assert "not in (date '" in rendered, rendered
    rendered = BaseDialect().render_expr(
        BuildComparison(
            left=env.materialize_for_select().concepts["order_id"],
            operator=ComparisonOperator.IN,
            right=wrapper,
        )
    )
    assert "in (date '" in rendered, rendered


def test_is_not_null():
    _, parsed = parse_text(
        "const order_id <- 4; SELECT order_id  WHERE order_id is not null;"
    )
    query = parsed[-1]
    right = query.where_clause.conditional.right
    assert isinstance(right, MagicConstants), type(right)
    rendered = BaseDialect().render_expr(right)
    assert rendered == "null"


def test_sort():
    _, parsed = parse_text(
        "const order_id <- 4; SELECT order_id  ORDER BY order_id desc;"
    )

    _, parsed = parse_text(
        "const order_id <- 4; SELECT order_id  ORDER BY order_id DESC;"
    )
    _, parsed = parse_text(
        "const order_id <- 4; SELECT order_id  ORDER BY order_id DesC;"
    )


def test_arg_to_datatype():
    assert arg_to_datatype(1.00) == DataType.FLOAT
    assert arg_to_datatype("test") == DataType.STRING


def test_argument_to_purpose(test_environment: Environment):
    assert (
        argument_to_purpose(
            1.00,
        )
        == Purpose.CONSTANT
    )
    assert argument_to_purpose("test") == Purpose.CONSTANT
    assert argument_to_purpose(test_environment.concepts["order_id"]) == Purpose.KEY
    assert (
        function_args_to_output_purpose(
            [
                "test",
                1.00,
            ],
            test_environment,
        )
        == Purpose.CONSTANT
    )
    assert (
        function_args_to_output_purpose(
            ["test", 1.00, test_environment.concepts["order_id"]], test_environment
        )
        == Purpose.PROPERTY
    )
    unnest_env, parsed = parse_text("const random <- unnest([1,2,3,4]);")
    assert (
        function_args_to_output_purpose(
            [unnest_env.concepts["random"]], test_environment
        )
        == Purpose.PROPERTY
    )


def test_show(test_environment):
    _, parsed = parse_text(
        "const order_id <- 4; SHOW SELECT order_id  WHERE order_id is not null;"
    )
    query = parsed[-1]
    assert isinstance(query, ShowStatement)
    assert (
        query.content.output_components[0].address
        == test_environment.concepts["order_id"].address
    )


def test_conditional(test_environment):
    _, parsed = parse_text(
        "const order_id <- 4; SELECT order_id  WHERE order_id =4 and order_id = 10;"
    )
    query = parsed[-1]
    assert isinstance(query, SelectStatement)
    assert query.where_clause.conditional.operator == BooleanOperator.AND


def test_as_transform(test_environment):
    _, parsed = parse_text("const order_id <- 4; SELECT order_id as new_order_id;")
    query = parsed[-1]
    assert isinstance(query, SelectStatement)


def test_bq_address():
    _, parsed = parse_text(
        """key user_pseudo_id int;
key event_time int;
property event_time.event_date string;

datasource fundiverse (
    event_date: event_date,
    user_pseudo_id: user_pseudo_id,
    event_time: event_time,
)
grain (event_time)
address `preqldata.analytics_411641820.events_*`
;"""
    )
    query = parsed[-1]
    assert query.address.quoted is True
    assert query.address.location == "preqldata.analytics_411641820.events_*"


def test_purpose_and_keys():
    env, parsed = parse_text(
        """key id int;
property id.name string;

auto name_alphabetical <- row_number id order by name asc;


select
    id,
    name,
    row_number id order by name asc -> name_alphabetical_2
    ;
"""
    )

    for name in [
        "name_alphabetical",
    ]:
        assert f"local.{name}" in env.concepts
        assert env.concepts[name].purpose == Purpose.PROPERTY
        assert env.concepts[name].keys == {env.concepts["id"].address}


def test_purpose_and_derivation():
    env, parsed = parse_text(
        """key id int;
key other_id int;
property <id, other_id>.join_id <- id*10+other_id;


select 
    join_id
;
"""
    )

    for name in ["join_id"]:
        assert env.concepts[name].purpose == Purpose.PROPERTY
        assert env.concepts[name].keys == {
            env.concepts["id"].address,
            env.concepts["other_id"].address,
        }


def test_output_purpose():
    env, parsed = parse_text(
        """key id int;
property id.name string;

auto name_alphabetical <- row_number id order by name asc;


rowset test<- select
    name,
    row_number id order by name asc -> name_alphabetical_2
    ;

auto test_name_count <- count(test.name);

select 
    test_name_count;
"""
    )
    # assert output_purpose == Purpose.METRIC
    for name in ["test_name_count"]:
        assert env.concepts[name].purpose == Purpose.METRIC


def test_between():
    _, parsed = parse_text(
        "const order_id <- 4; SELECT order_id  WHERE order_id BETWEEN 3 and 5;"
    )
    query: ProcessedQuery = parsed[-1]
    left = query.where_clause.conditional.left
    assert isinstance(
        left,
        Comparison,
    ), type(left)
    assert left.operator == ComparisonOperator.GTE
    assert left.right == 3

    right = query.where_clause.conditional.right
    assert isinstance(
        right,
        Comparison,
    ), type(right)
    assert right.operator == ComparisonOperator.LTE
    assert right.right == 5


def test_the_comments():
    _, parsed = parse_text(
        """const
         # comment here?
           order_id <- 4; SELECT 
        # TOOD - add in more columns?
        order_id   # this is the order id
        WHERE 
        # order_id should not be null
        order_id
        # in this comp
          is not 
        null; # nulls are the worst
        
        """
    )
    query = parsed[-1]
    right = query.where_clause.conditional.right
    assert isinstance(right, MagicConstants), type(right)
    rendered = BaseDialect().render_expr(right)
    assert rendered == "null"


def test_the_comment_multiline():
    env, parsed = parse_text(
        """const
         # comment here?
           order_id <- 4;  # this is the order id
        # order ids are important
           
           SELECT 
        # TOOD - add in more columns?
        order_id  
        WHERE 
        # order_id should not be null
        order_id
        # in this comp
          is not 
        null; # nulls are the worst
        
        """
    )
    parsed[-1]
    assert env.concepts["order_id"].metadata.description is not None
    assert " this is the order id" in env.concepts["order_id"].metadata.description
    assert " order ids are important" in env.concepts["order_id"].metadata.description


def test_the_comment_multiline_enter():
    # we should not associate it as a description if there is a newline
    env, parsed = parse_text(
        """const
         # comment here?
           order_id <- 4;  
        # this is the order id
        # order ids are important
           
           SELECT 
        # TOOD - add in more columns?
        order_id  
        WHERE 
        # order_id should not be null
        order_id
        # in this comp
          is not 
        null; # nulls are the worst
        
        """
    )
    assert env.concepts["order_id"].metadata.description is None


def test_purpose_nesting():
    env, parsed = parse_text(
        """key year int;
"""
    )

    env2: Environment = Environment()
    env2.add_import("dates", env)

    env2, _ = parse_text(
        """
property <dates.year>.generation <-
CASE WHEN dates.year BETWEEN 1883 AND 1900 THEN 'Lost Generation'
        WHEN dates.year BETWEEN 1901 AND 1927 THEN 'The Greatest Generation'
        WHEN dates.year BETWEEN 1928 AND 1945 THEN 'The Silent Generation'
        WHEN dates.year BETWEEN 1946 AND 1964 THEN 'Baby Boomer'
        WHEN dates.year BETWEEN 1965 AND 1980 THEN 'Generation X'
        WHEN dates.year BETWEEN 1981 AND 1996 THEN 'Millennials'
        WHEN dates.year BETWEEN 1997 AND 2012 THEN 'Generation Z'
        ELSE 'Unknown'
    END;
    """,
        env2,
    )

    assert env2.concepts["dates.generation"].purpose == Purpose.PROPERTY


def test_rawsql():
    env, parsed = parse_text(
        """
raw_sql('''select 1''');

select 1 as test;

"""
    )
    assert parsed[0].text == "select 1"


def test_circular_aliasing():
    from trilogy.hooks.query_debugger import DebuggingHook

    executor = Dialects.DUCK_DB.default_executor(hooks=[DebuggingHook()])
    test_case = """key composite_id string;

property composite_id.first <- split(composite_id, '-')[1];
property composite_id.second <- split(composite_id, '-')[2];

auto composite_id_alt <- concat(first, '-', second);

merge composite_id_alt into composite_id;

datasource random (
    first:first,
    second:second
)
grain (composite_id)
query '''
select '123' as first, 'abc' as second
'''
;

datasource metrics (
    composite_id: composite_id,
)
grain (composite_id)
query '''
select '123-abc' as composite_id
'''
;


select first, second;


"""
    executor.parse_text(test_case)

    results = executor.execute_text(test_case)[0].fetchall()

    assert results == [("123", "abc")]


def test_circular_aliasing_inverse():
    from trilogy.hooks.query_debugger import DebuggingHook

    executor = Dialects.DUCK_DB.default_executor(hooks=[DebuggingHook()])
    test_case = """key composite_id string;

property composite_id.first <- split(composite_id, '-')[1];
property composite_id.second <- split(composite_id, '-')[2];

auto composite_id_alt <- concat(first, '-', second);

merge composite_id_alt into composite_id;


datasource metrics (
    first:first,
    second:second
)
grain (composite_id)
query '''
select '123' as first, 'abc' as second
'''
;


select composite_id;


"""
    executor.parse_text(test_case)

    assert (
        "local.composite_id_alt"
        in executor.environment.concepts["local.composite_id"].pseudonyms
    )
    assert (
        "local.composite_id"
        in executor.environment.alias_origin_lookup["local.composite_id_alt"].pseudonyms
    )
    results = executor.execute_text(test_case)[0].fetchall()

    assert results == [("123-abc",)]


def test_map_definition():
    env, parsed = parse_text(
        """
key id int;
property id.labels map<string, int>;

"""
    )
    assert env.concepts["labels"].datatype.key_type == DataType.STRING
    assert env.concepts["labels"].datatype.value_type == DataType.INTEGER


def test_map_concept_definition():
    env, parsed = parse_text(
        """
key id int;
property id.label string;
key map_store map<id, string>;

"""
    )
    assert env.concepts["map_store"].datatype.key_data_type == DataType.INTEGER
    assert env.concepts["map_store"].datatype.value_data_type == DataType.STRING


def test_map_string_access():
    env, parsed = parse_text(
        """
const labels <- { 'a': 1, 'b': 2, 'c': 3 };


select
    labels['a'] as label_a,
;

"""
    )
    assert env.concepts["labels"].datatype.key_type == DataType.STRING
    assert env.concepts["labels"].datatype.value_type == DataType.INTEGER


def test_empty_string():
    env, parsed = parse_text(
        """
const labels <- '';


"""
    )
    assert env.concepts["labels"].datatype == DataType.STRING


def test_struct_attr_access():
    text = """
const labels <- struct(1->a, 2->b, 3->c);


select
    labels,
    labels.a as label_a
;

"""
    env, parsed = parse_text(text)
    assert env.concepts["labels"].datatype.fields_map["a"] == 1

    assert env.concepts["labels.a"].concept_arguments[0].name == "labels"

    results = Dialects.DUCK_DB.default_executor().execute_text(text)[0]

    assert results.fetchall()[0] == (
        {"a": 1, "b": 2, "c": 3},
        1,
    )


def test_datasource_colon():
    text = """
key x int;
key y int;

datasource test (
x:x,
y:y)
grain(x)
address `abc:def`
;


select x;
"""
    env, parsed = parse_text(text)

    results = Dialects.DUCK_DB.default_executor().generate_sql(text)[0]

    assert '"abc:def" as "test"' in results

    text = """
key x int;
key y int;

datasource test (
x:x,
y:y)
grain(x)
address abcdef
;


select x;
"""
    env, parsed = parse_text(text)

    results = Dialects.DUCK_DB.default_executor().generate_sql(text)[0]

    assert '"abcdef" as "test"' in results, results

    text = """
key x int;
key y int;

datasource test (
x,
y)
grain(x)
address abcdef
;


select x;
"""
    env, parsed = parse_text(text)

    results = Dialects.DUCK_DB.default_executor().generate_sql(text)[0]

    assert '"abcdef" as "test"' in results, results


def test_datasource_where_equivalent():
    text = """
key x int;
key y int;

datasource test (
x:x,
y:~y)
grain(x)
complete where y > 10
address `abc:def`
;


"""
    env, parsed = parse_text(text)

    ds = parsed[-1]
    assert ds.non_partial_for.conditional.right == 10


def test_datasource_quoted():
    text = """
key x int;
key y int;

datasource test (
`x 2`:x,
`y`:~y)
grain(x)
complete where y > 10
address `abc:def`
;


"""
    env, parsed = parse_text(text)

    ds = parsed[-1]
    assert ds.columns[0].alias == "x 2"


def test_datasource_from_persist():
    text = """
key x int;
key y int;

datasource test (
x:x,
y:y)
grain(x)
address `abc:def`
;

persist alias into tbl_alias from
select
x,
y
where y>10;



"""
    env, parsed = parse_text(text)

    ds: Datasource = parsed[-1].datasource
    assert ds.non_partial_for.conditional.right == 10
    assert not ds.where


def test_filter_concise():
    text = """
key x int;
key y int;

datasource test (
x:x,
y:y)
grain(x)
address `abc:def`
;

auto filtered_test <- x ? y > 10;

select filtered_test;
"""
    env, parsed = parse_text(text)

    results = Dialects.DUCK_DB.default_executor().generate_sql(text)[0]

    assert "filtered_test" in results, results


def test_unnest_parsing():
    x = """
key scalar int;    
property scalar.int_array list<int>;

key split <- unnest(int_array);

datasource avalues (
    int_array: int_array,
	scalar: scalar
    ) 
grain (scalar) 
query '''(
select [1,2,3,4] as int_array, 2 as scalar
)''';
"""

    env, parsed = parse_text(x)
    assert env.concepts["split"].datatype == DataType.INTEGER


def test_non_file_imports():

    env = Environment(
        config=EnvironmentOptions(
            import_resolver=DictImportResolver(
                content={
                    "test": """
import test_dep as test_dep;
key x int;

datasource test (
x: x)
grain(x)
query '''
select 1 as x
union all
select 11 as x
''';
""",
                    "test_dep": """
key x int;
""",
                }
            )
        )
    )
    assert isinstance(env.config.import_resolver, DictImportResolver)
    env.parse(
        """
import std.geography;
import test;

key fun_lat float::latitude;
               
select x % 10 -> x_mod_10;
               
            
"""
    )


def test_is_atom():
    env = Environment()

    env.parse(
        """
key x int;
auto x_sum <- sum(x);
"""
    )

    assert not atom_is_relevant(
        Comparison(left=env.concepts["x_sum"], right=0, operator=ComparisonOperator.GT),
        [env.concepts["x"]],
        env,
    )
    assert atom_is_relevant(
        Comparison(left=env.concepts["x"], right=0, operator=ComparisonOperator.GT),
        [env.concepts["x_sum"]],
        env,
    )

    assert not atom_is_relevant(
        Conditional(left=env.concepts["x_sum"], right=0, operator=BooleanOperator.AND),
        [env.concepts["x"]],
        env,
    )
    assert atom_is_relevant(
        Conditional(left=env.concepts["x"], right=0, operator=BooleanOperator.AND),
        [env.concepts["x_sum"]],
        env,
    )
    assert not atom_is_relevant(
        SubselectComparison(
            left=env.concepts["x_sum"],
            right=env.concepts["x"],
            operator=ComparisonOperator.IN,
        ),
        [env.concepts["x_sum"]],
        env,
    )



from pathlib import Path

from trilogy import Dialects, parse
from trilogy.core.models.environment import Environment
from trilogy.dialect.metadata import MockResult, MockResultRow


def test_file_parsing():
    directory = Path(__file__).parent
    target = directory / "test_env.preql"
    parsed = Dialects.DUCK_DB.default_executor(
        environment=Environment(working_path=directory)
    ).parse_file(target)
    assert len(list(parsed)) == 1


def test_can_handle_everything():
    env = Environment(working_path=Path(__file__).parent)
    env, queries = parse(
        """
import test_env as test_env;

key x int;

merge test_env.id into x;
                         
RAW_SQL('''select 1 ''');
""",
        environment=env,
    )
    execs = Dialects.DUCK_DB.default_executor(environment=env)
    for q in queries:
        execs.execute_query(q)


def test_mock_result():
    result = MockResult(
        columns=["a", "b", "c"],
        values=[
            MockResultRow({"a": 1, "b": 2, "c": 3}),
            MockResultRow({"a": 4, "b": 5, "c": 6}),
        ],
    )
    assert result.columns == ["a", "b", "c"]

    for row in result:
        assert isinstance(row, MockResultRow)
        assert set(result.keys()) == {"a", "b", "c"}
        assert all(isinstance(v, int) for v in row.values())

    result = MockResult(
        columns=["a", "b", "c"],
        values=[
            MockResultRow({"a": 1, "b": 2, "c": 3}),
            MockResultRow({"a": 4, "b": 5, "c": 6}),
        ],
    )
    assert result.columns == ["a", "b", "c"]

    x = result.fetchone()
    assert isinstance(x, MockResultRow)
    assert x["a"] == 1
    r2 = result.fetchmany(10)
    assert len(r2) == 1



import random
from dataclasses import dataclass, field
from enum import Enum
from logging import getLogger

logger = getLogger("trilogy")

DEFAULT_NAMESPACE = "local"

RECURSIVE_GATING_CONCEPT = "_terminal"

VIRTUAL_CONCEPT_PREFIX = "_virt"

ENV_CACHE_NAME = ".preql_cache.json"


class MagicConstants(Enum):
    NULL = "null"
    LINE_SEPARATOR = "\n"


NULL_VALUE = MagicConstants.NULL


@dataclass
class Optimizations:
    predicate_pushdown: bool = True
    datasource_inlining: bool = True
    constant_inlining: bool = True
    constant_inline_cutoff: int = 10
    direct_return: bool = True


@dataclass
class Comments:
    """Control what is placed in CTE comments"""

    show: bool = False
    basic: bool = True
    joins: bool = False
    nullable: bool = True
    partial: bool = True
    source_map: bool = False


@dataclass
class Rendering:
    """Control how the SQL is rendered"""

    parameters: bool = True
    concise: bool = False


@dataclass
class Parsing:
    """Control Parsing"""

    strict_name_shadow_enforcement: bool = False
    select_as_definition: bool = True


# TODO: support loading from environments
@dataclass
class Config:
    strict_mode: bool = True
    human_identifiers: bool = True
    randomize_cte_names: bool = False
    validate_missing: bool = True
    comments: Comments = field(default_factory=Comments)
    optimizations: Optimizations = field(default_factory=Optimizations)
    rendering: Rendering = field(default_factory=Rendering)
    parsing: Parsing = field(default_factory=Parsing)

    @property
    def show_comments(self) -> bool:
        return self.comments.show

    def set_random_seed(self, seed: int):
        random.seed(seed)


CONFIG = Config()

CONFIG.set_random_seed(42)



from abc import ABC

from trilogy.constants import logger
from trilogy.core.models.execute import CTE, UnionCTE


class OptimizationRule(ABC):
    def optimize(
        self, cte: CTE | UnionCTE, inverse_map: dict[str, list[CTE | UnionCTE]]
    ) -> bool:
        raise NotImplementedError

    def log(self, message: str):
        logger.info(f"[Optimization][{self.__class__.__name__}] {message}")

    def debug(self, message: str):
        logger.debug(f"[Optimization][{self.__class__.__name__}] {message}")



from copy import deepcopy

from pytest import raises

from trilogy import Dialects, parse
from trilogy.core.enums import (
    BooleanOperator,
    ComparisonOperator,
    FunctionType,
    JoinType,
    Purpose,
)
from trilogy.core.models.author import (
    AggregateWrapper,
    Comparison,
    Concept,
    Conditional,
    Function,
    Grain,
    Parenthetical,
    RowsetItem,
    UndefinedConcept,
)
from trilogy.core.models.build import BuildGrain
from trilogy.core.models.core import (
    DataType,
    ListWrapper,
    TupleWrapper,
)
from trilogy.core.models.datasource import Address
from trilogy.core.models.environment import Environment
from trilogy.core.models.execute import (
    CTE,
    BaseJoin,
    CTEConceptPair,
    Join,
    QueryDatasource,
)
from trilogy.core.statements.author import SelectStatement


def test_cte_merge(test_environment, test_environment_graph):
    test_environment = test_environment.materialize_for_select()
    datasource = list(test_environment.datasources.values())[0]
    outputs = [c.concept for c in datasource.columns]
    output_map = {
        c.address: {
            datasource,
        }
        for c in outputs
    }
    a = CTE(
        name="test",
        output_columns=[outputs[0]],
        grain=BuildGrain(),
        source=QueryDatasource(
            input_concepts=[outputs[0]],
            output_concepts=[outputs[0]],
            datasources=[datasource],
            grain=BuildGrain(),
            joins=[],
            source_map={outputs[0].address: {datasource}},
        ),
        source_map={c.address: [datasource.identifier] for c in outputs},
    )
    b = CTE(
        name="testb",
        output_columns=outputs,
        grain=BuildGrain(),
        source=QueryDatasource(
            input_concepts=outputs,
            output_concepts=outputs,
            datasources=[datasource],
            grain=BuildGrain(),
            joins=[],
            source_map=output_map,
        ),
        source_map={c.address: [datasource.identifier] for c in outputs},
    )

    merged = a + b
    assert merged.output_columns == outputs

    assert "Target: Grain<Abstract>." in merged.comment


def test_concept(test_environment, test_environment_graph):
    test_concept: Concept = list(test_environment.concepts.values())[0]
    new = test_concept.with_namespace("test")
    assert (
        new.namespace == ("test" + "." + test_concept.namespace)
        if test_concept.namespace != "local"
        else "test"
    )


def test_concept_filter(test_environment, test_environment_graph):
    test_concept: Concept = list(test_environment.concepts.values())[0]
    new = test_concept.with_filter(
        Comparison(left=1, right=2, operator=ComparisonOperator.EQ)
    )
    new2 = test_concept.with_filter(
        Comparison(left=1, right=2, operator=ComparisonOperator.EQ)
    )

    assert new.name == new2.name != test_concept.name

    new3 = new.with_filter(Comparison(left=1, right=2, operator=ComparisonOperator.EQ))
    assert new3 == new


def test_conditional(test_environment, test_environment_graph):
    test_concept = list(test_environment.concepts.values())[-1]

    condition_a = Conditional(
        left=test_concept, right=test_concept, operator=BooleanOperator.AND
    )
    condition_b = Conditional(
        left=test_concept, right=test_concept, operator=BooleanOperator.AND
    )
    merged = condition_a + condition_b
    assert merged == condition_a

    test_concept_two = [
        x
        for x in test_environment.concepts.values()
        if x.address != test_concept.address
    ].pop()
    condition_c = Conditional(
        left=test_concept, right=test_concept_two, operator=BooleanOperator.AND
    )
    merged_two = condition_a + condition_c
    assert merged_two.left == condition_a, f"{str(merged_two.left), str(condition_a)}"
    assert merged_two.right == condition_c
    assert merged_two.operator == BooleanOperator.AND


def test_grain(test_environment):
    grains = [Grain()] * 3
    sum(grains)
    oid = test_environment.concepts["order_id"]
    pid = test_environment.concepts["product_id"]
    cid = test_environment.concepts["category_id"]
    cname = test_environment.concepts["category_name"]

    x = Grain(components=[oid, pid])
    y = Grain(components=[pid, cid])
    z = Grain(components=[cid])
    z2 = Grain.from_concepts([cid, cname], environment=test_environment)

    assert x.intersection(y) == Grain(components=[pid])
    assert x.union(y) == Grain(components=[oid, pid, cid])

    assert z.isdisjoint(x)
    assert z.issubset(y)

    assert (
        z2 == z
    ), f"Property should be removed from grain ({z.components}) vs {z2.components}"


def test_select(test_environment: Environment):
    oid = test_environment.concepts["order_id"]
    pid = test_environment.concepts["product_id"]
    cid = test_environment.concepts["category_id"]
    cname = test_environment.concepts["category_name"]
    x = SelectStatement(
        selection=[oid, pid, cid, cname], grain=Grain(components=[oid, pid, cid])
    )
    ds = x.to_datasource(
        test_environment.namespace,
        "test",
        address=Address(location="test"),
        environment=test_environment,
    )

    assert ds.grain.components == Grain(components=[oid, pid, cid]).components


def test_undefined(test_environment: Environment):
    UndefinedConcept(address="test.test")


def test_base_join(test_environment: Environment):
    exc: SyntaxError | None = None
    test_environment = test_environment.materialize_for_select()
    try:
        BaseJoin(
            left_datasource=test_environment.datasources["revenue"],
            right_datasource=test_environment.datasources["revenue"],
            concepts=[test_environment.concepts["product_id"]],
            join_type=JoinType.RIGHT_OUTER,
        )
    except Exception as exc2:
        exc = exc2
        pass
    assert isinstance(exc, SyntaxError)

    x = BaseJoin(
        left_datasource=test_environment.datasources["revenue"],
        right_datasource=test_environment.datasources["products"],
        concepts=[
            test_environment.concepts["product_id"],
            # test_environment.concepts["category_name"],
        ],
        join_type=JoinType.RIGHT_OUTER,
    )

    assert x.concepts == [test_environment.concepts["product_id"]]


def test_comparison():
    with raises(SyntaxError):
        Comparison(left=1, right="abc", operator=ComparisonOperator.EQ)

    # this should not error
    Comparison(left=1, right=[1, 2, 3], operator=ComparisonOperator.IN)

    Comparison(
        left=1,
        right=ListWrapper([1, 2, 3], type=DataType.INTEGER),
        operator=ComparisonOperator.IN,
    )


def test_comparison_invalid_type():
    try:
        Comparison(left=1, right="abc", operator=ComparisonOperator.EQ)
    except Exception as exc:
        assert isinstance(exc, SyntaxError)


def test_join(test_environment: Environment):
    test_environment = test_environment.materialize_for_select()
    datasource = list(test_environment.datasources.values())[0]
    outputs = [c.concept for c in datasource.columns]
    output_map = {
        c.address: {
            datasource,
        }
        for c in outputs
    }
    a = CTE(
        name="test",
        output_columns=[outputs[0]],
        grain=BuildGrain(),
        source=QueryDatasource(
            input_concepts=[outputs[0]],
            output_concepts=[outputs[0]],
            datasources=[datasource],
            grain=BuildGrain(),
            joins=[],
            source_map={outputs[0].address: {datasource}},
        ),
        source_map={c.address: [datasource.identifier] for c in outputs},
    )

    b = CTE(
        name="testb",
        output_columns=outputs,
        grain=BuildGrain(),
        source=QueryDatasource(
            input_concepts=outputs,
            output_concepts=outputs,
            datasources=[datasource],
            grain=BuildGrain(),
            joins=[],
            source_map=output_map,
        ),
        source_map={c.address: [datasource.identifier] for c in outputs},
    )
    test = Join(
        left_cte=a,
        right_cte=b,
        joinkey_pairs=[
            CTEConceptPair(left=x, right=x, existing_datasource=a.source, cte=a)
            for x in outputs
        ],
        jointype=JoinType.RIGHT_OUTER,
    )

    assert (
        str(test)
        == "right outer join testb on test.local.product_id=local.product_id,test.local.category_id=local.category_id"
    ), str(test)


def test_concept_address_in_check():
    target = Concept(
        name="test",
        datatype="int",
        purpose=Purpose.CONSTANT,
        grain=Grain(),
    )
    assert target.address == "local.test"
    x = [target]

    assert "local.test" in x


def test_rowset_with_filter_derivation():
    env, statements = parse(
        """
key x int;

datasource test (
x:x
)
grain (x)
address test
;

with greater_than_ten as
select x
where 
x >10;

auto avg_greater_ten <- avg(greater_than_ten.x) by *;

select avg_greater_ten;

"""
    )

    lineage = env.concepts["avg_greater_ten"].lineage
    assert isinstance(lineage, AggregateWrapper)
    assert isinstance(
        env.concepts[lineage.function.concept_arguments[0].address].lineage, RowsetItem
    )


def test_tuple_clone():
    x = TupleWrapper([1, 2, 3], type=DataType.INTEGER)
    y = deepcopy(x)
    assert y == x


def test_parenthetical(test_environment: Environment):
    x = Parenthetical(content=TupleWrapper([1, 2, 3], type=DataType.INTEGER))

    assert x.concept_arguments == []

    x = Parenthetical(
        content=test_environment.concepts["order_id"], type=DataType.INTEGER
    )
    # return concept if it's a concept
    assert x.concept_arguments == [test_environment.concepts["order_id"]]
    function = Function(
        operator=FunctionType.COUNT,
        output_datatype=DataType.INTEGER,
        output_purpose=Purpose.METRIC,
        arguments=[test_environment.concepts["order_id"]],
    )
    assert function.concept_arguments == [
        test_environment.concepts["order_id"].reference
    ]
    x = Parenthetical(content=function)
    # else pass through parent
    assert x.concept_arguments == [test_environment.concepts["order_id"].reference]

    merged = x + x
    assert isinstance(merged, Conditional)


def test_datasource_grain_application():
    env, statements = parse(
        """
key x string;
property x.part_1 string;
property x.part_2 string;
auto _x <- concat(x.part_1, '.', x.part_2);

merge _x into x;

datasource xes (
    x:x,
    )
grain (x)
address x_table;

datasource parts (
    part_1: part_1,
    part_2:part_2,
    concat(x.part_1, '.', x.part_2): x
    )
grain (part_1, part_2)
address parts;



"""
    )

    test_grain = Grain.from_concepts(
        concepts=["x", "part_1", "part_2"], environment=env
    )
    assert test_grain.components == {
        "local.x",
    }


def test_datasource_merge_generation():
    env, statements = parse(
        """
key x string;
property x.part_1 string;
property x.part_2 string;
auto _x <- concat(part_1, '.', part_2);

merge _x into x;

datasource xes (
    x:x,
    )
grain (x)
address x_table;

datasource parts (
    part_1: part_1,
    part_2:part_2,
    concat(part_1, '.', part_2): x
    )
grain (part_1, part_2)
address parts;


select x,
    part_1,
    part_2
;


"""
    )

    q = Dialects.DUCK_DB.default_executor(environment=env).generate_sql(statements[-1])[
        0
    ]

    assert "JOIN" not in q, q



from __future__ import annotations

from collections import defaultdict
from typing import Dict, List, Optional, Set, Union

from pydantic import (
    BaseModel,
    Field,
    ValidationInfo,
    computed_field,
    field_validator,
    model_validator,
)

from trilogy.constants import (
    CONFIG,
    DEFAULT_NAMESPACE,
    RECURSIVE_GATING_CONCEPT,
    MagicConstants,
    logger,
)
from trilogy.core.constants import CONSTANT_DATASET
from trilogy.core.enums import (
    ComparisonOperator,
    Derivation,
    FunctionClass,
    FunctionType,
    JoinType,
    Modifier,
    Purpose,
    SourceType,
)
from trilogy.core.exceptions import InvalidSyntaxException
from trilogy.core.models.build import (
    BuildCaseElse,
    BuildCaseWhen,
    BuildComparison,
    BuildConcept,
    BuildConditional,
    BuildDatasource,
    BuildExpr,
    BuildFunction,
    BuildGrain,
    BuildOrderBy,
    BuildParamaterizedConceptReference,
    BuildParenthetical,
    BuildRowsetItem,
    DataType,
    LooseBuildConceptList,
)
from trilogy.core.models.datasource import Address
from trilogy.core.utility import safe_quote
from trilogy.utility import unique

LOGGER_PREFIX = "[MODELS_EXECUTE]"

DATASOURCE_TYPES = (BuildDatasource, BuildDatasource)


class InlinedCTE(BaseModel):
    original_alias: str
    new_alias: str
    new_base: str


class CTE(BaseModel):
    name: str
    source: "QueryDatasource"
    output_columns: List[BuildConcept]
    source_map: Dict[str, list[str]]
    grain: BuildGrain
    base: bool = False
    group_to_grain: bool = False
    existence_source_map: Dict[str, list[str]] = Field(default_factory=dict)
    parent_ctes: List[Union["CTE", "UnionCTE"]] = Field(default_factory=list)
    joins: List[Union["Join", "InstantiatedUnnestJoin"]] = Field(default_factory=list)
    condition: Optional[
        Union[BuildComparison, BuildConditional, BuildParenthetical]
    ] = None
    partial_concepts: List[BuildConcept] = Field(default_factory=list)
    nullable_concepts: List[BuildConcept] = Field(default_factory=list)
    join_derived_concepts: List[BuildConcept] = Field(default_factory=list)
    hidden_concepts: set[str] = Field(default_factory=set)
    order_by: Optional[BuildOrderBy] = None
    limit: Optional[int] = None
    base_name_override: Optional[str] = None
    base_alias_override: Optional[str] = None
    inlined_ctes: dict[str, InlinedCTE] = Field(default_factory=dict)

    @field_validator("join_derived_concepts")
    def validate_join_derived_concepts(cls, v):
        if len(v) > 1:
            raise NotImplementedError(
                "Multiple join derived concepts not yet supported."
            )
        return unique(v, "address")

    @property
    def identifier(self):
        return self.name

    @property
    def safe_identifier(self):
        return self.name

    @computed_field  # type: ignore
    @property
    def output_lcl(self) -> LooseBuildConceptList:
        return LooseBuildConceptList(concepts=self.output_columns)

    @field_validator("output_columns")
    def validate_output_columns(cls, v):
        return unique(v, "address")

    @property
    def comment(self) -> str:
        base = f"Target: {str(self.grain)}. Group: {self.group_to_grain}"
        base += f" Source: {self.source.source_type}."
        if self.parent_ctes:
            base += f" References: {', '.join([x.name for x in self.parent_ctes])}."
        if self.joins and CONFIG.comments.joins:
            base += f"\n-- Joins: {', '.join([str(x) for x in self.joins])}."
        if self.partial_concepts and CONFIG.comments.partial:
            base += (
                f"\n-- Partials: {', '.join([str(x) for x in self.partial_concepts])}."
            )
        if CONFIG.comments.source_map:
            base += f"\n-- Source Map: {self.source_map}."
        base += f"\n-- Output: {', '.join([str(x) for x in self.output_columns])}."
        if self.source.input_concepts:
            base += f"\n-- Inputs: {', '.join([str(x) for x in self.source.input_concepts])}."
        if self.hidden_concepts:
            base += f"\n-- Hidden: {', '.join([str(x) for x in self.hidden_concepts])}."
        if self.nullable_concepts and CONFIG.comments.nullable:
            base += (
                f"\n-- Nullable: {', '.join([str(x) for x in self.nullable_concepts])}."
            )
        base += "\n"
        return base

    def inline_parent_datasource(
        self, parent: "CTE", force_group: bool = False
    ) -> bool:
        qds_being_inlined = parent.source
        ds_being_inlined = qds_being_inlined.datasources[0]
        if not isinstance(ds_being_inlined, DATASOURCE_TYPES):
            return False
        if any(
            [
                x.safe_identifier == ds_being_inlined.safe_identifier
                for x in self.source.datasources
            ]
        ):
            return False
        self.source.datasources = [
            ds_being_inlined,
            *[
                x
                for x in self.source.datasources
                if x.safe_identifier != qds_being_inlined.safe_identifier
            ],
        ]
        # need to identify this before updating joins
        if self.base_name == parent.name:
            self.base_name_override = ds_being_inlined.safe_location
            self.base_alias_override = ds_being_inlined.safe_identifier

        # if we have a join to the parent, we need to remove it
        for join in self.joins:
            if isinstance(join, InstantiatedUnnestJoin):
                continue
            if (
                join.left_cte
                and join.left_cte.safe_identifier == parent.safe_identifier
            ):
                join.inline_cte(parent)
            if join.joinkey_pairs:
                for pair in join.joinkey_pairs:
                    if pair.cte and pair.cte.safe_identifier == parent.safe_identifier:
                        join.inline_cte(parent)
            if join.right_cte.safe_identifier == parent.safe_identifier:
                join.inline_cte(parent)
        for k, v in self.source_map.items():
            if isinstance(v, list):
                self.source_map[k] = [
                    (
                        ds_being_inlined.safe_identifier
                        if x == parent.safe_identifier
                        else x
                    )
                    for x in v
                ]
            elif v == parent.safe_identifier:
                self.source_map[k] = [ds_being_inlined.safe_identifier]
        for k, v in self.existence_source_map.items():
            if isinstance(v, list):
                self.existence_source_map[k] = [
                    (
                        ds_being_inlined.safe_identifier
                        if x == parent.safe_identifier
                        else x
                    )
                    for x in v
                ]
            elif v == parent.safe_identifier:
                self.existence_source_map[k] = [ds_being_inlined.safe_identifier]
        # zip in any required values for lookups
        for k in ds_being_inlined.output_lcl.addresses:
            if k in self.source_map and self.source_map[k]:
                continue
            self.source_map[k] = [ds_being_inlined.safe_identifier]
        self.parent_ctes = [
            x for x in self.parent_ctes if x.safe_identifier != parent.safe_identifier
        ]
        if force_group:
            self.group_to_grain = True
        self.inlined_ctes[ds_being_inlined.safe_identifier] = InlinedCTE(
            original_alias=parent.name,
            new_alias=ds_being_inlined.safe_identifier,
            new_base=ds_being_inlined.safe_location,
        )
        return True

    def __add__(self, other: "CTE" | "UnionCTE"):
        if isinstance(other, UnionCTE):
            raise ValueError("cannot merge CTE and union CTE")
        logger.info('Merging two copies of CTE "%s"', self.name)
        if not self.grain == other.grain:
            error = (
                "Attempting to merge two ctes of different grains"
                f" {self.name} {other.name} grains {self.grain} {other.grain}| {self.group_to_grain} {other.group_to_grain}| {self.output_lcl} {other.output_lcl}"
            )
            raise ValueError(error)
        if not self.condition == other.condition:
            error = (
                "Attempting to merge two ctes with different conditions"
                f" {self.name} {other.name} conditions {self.condition} {other.condition}"
            )
            raise ValueError(error)
        mutually_hidden = set()
        for concept in self.hidden_concepts:
            if concept in other.hidden_concepts:
                mutually_hidden.add(concept)
        self.partial_concepts = unique(
            self.partial_concepts + other.partial_concepts, "address"
        )
        self.parent_ctes = merge_ctes(self.parent_ctes + other.parent_ctes)

        self.source_map = {**self.source_map, **other.source_map}

        self.output_columns = unique(
            self.output_columns + other.output_columns, "address"
        )
        self.joins = unique(self.joins + other.joins, "unique_id")
        self.partial_concepts = unique(
            self.partial_concepts + other.partial_concepts, "address"
        )
        self.join_derived_concepts = unique(
            self.join_derived_concepts + other.join_derived_concepts, "address"
        )

        self.source.source_map = {**self.source.source_map, **other.source.source_map}
        self.source.output_concepts = unique(
            self.source.output_concepts + other.source.output_concepts, "address"
        )
        self.nullable_concepts = unique(
            self.nullable_concepts + other.nullable_concepts, "address"
        )
        self.hidden_concepts = mutually_hidden
        self.existence_source_map = {
            **self.existence_source_map,
            **other.existence_source_map,
        }
        self.inlined_ctes = {
            **self.inlined_ctes,
            **other.inlined_ctes,
        }

        return self

    @property
    def relevant_base_ctes(self):
        return self.parent_ctes

    @property
    def is_root_datasource(self) -> bool:
        return (
            len(self.source.datasources) == 1
            and isinstance(self.source.datasources[0], DATASOURCE_TYPES)
            and not self.source.datasources[0].name == CONSTANT_DATASET
        )

    @property
    def base_name(self) -> str:
        if self.base_name_override:
            return self.base_name_override
        # if this cte selects from a single datasource, select right from it
        if self.is_root_datasource:
            return self.source.datasources[0].safe_location

        # if we have multiple joined CTEs, pick the base
        # as the root
        elif len(self.source.datasources) == 1 and len(self.parent_ctes) == 1:
            return self.parent_ctes[0].name
        elif self.relevant_base_ctes:
            return self.relevant_base_ctes[0].name
        return self.source.name

    @property
    def quote_address(self) -> bool:
        if self.is_root_datasource:
            root = self.source.datasources[0]
            if isinstance(root, BuildDatasource) and isinstance(root.address, Address):
                return not root.address.is_query
            return True
        elif not self.source.datasources:
            return False
        base = self.source.datasources[0]
        if isinstance(base, BuildDatasource):
            if isinstance(base.address, Address):
                return not base.address.is_query
            return True
        return True

    @property
    def base_alias(self) -> str:
        if self.base_alias_override:
            return self.base_alias_override
        if self.is_root_datasource:
            return self.source.datasources[0].identifier
        elif self.relevant_base_ctes:
            return self.relevant_base_ctes[0].name
        elif self.parent_ctes:
            return self.parent_ctes[0].name
        return self.name

    def get_concept(self, address: str) -> BuildConcept | None:
        for cte in self.parent_ctes:
            if address in cte.output_columns:
                match = [x for x in cte.output_columns if x.address == address].pop()
                if match:
                    return match

        for array in [self.source.input_concepts, self.source.output_concepts]:
            match_list = [x for x in array if x.address == address]
            if match_list:
                return match_list.pop()
        match_list = [x for x in self.output_columns if x.address == address]
        if match_list:
            return match_list.pop()
        return None

    def get_alias(self, concept: BuildConcept, source: str | None = None) -> str:
        for cte in self.parent_ctes:
            if concept.address in cte.output_columns:
                if source and source != cte.name:
                    continue
                return concept.safe_address

        try:
            source = self.source.get_alias(concept, source=source)

            if not source:
                raise ValueError("No source found")
            return source
        except ValueError as e:
            return f"INVALID_ALIAS: {str(e)}"

    @property
    def group_concepts(self) -> List[BuildConcept]:
        def check_is_not_in_group(c: BuildConcept):

            if len(self.source_map.get(c.address, [])) > 0:
                return False
            if c.derivation == Derivation.ROWSET:
                assert isinstance(c.lineage, BuildRowsetItem)
                return check_is_not_in_group(c.lineage.content)
            if c.derivation == Derivation.CONSTANT:
                return True
            if (
                c.purpose == Purpose.CONSTANT
                and isinstance(c.lineage, BuildFunction)
                and c.lineage.operator in FunctionClass.AGGREGATE_FUNCTIONS.value
            ):
                return True

            if c.derivation == Derivation.BASIC and c.lineage:
                if all([check_is_not_in_group(x) for x in c.lineage.concept_arguments]):
                    return True
                if (
                    isinstance(c.lineage, BuildFunction)
                    and c.lineage.operator == FunctionType.GROUP
                ):
                    return check_is_not_in_group(c.lineage.concept_arguments[0])
                return False
            if c.purpose == Purpose.METRIC:
                return True

            return False

        return (
            unique(
                [c for c in self.output_columns if not check_is_not_in_group(c)],
                "address",
            )
            if self.group_to_grain
            else []
        )

    @property
    def render_from_clause(self) -> bool:
        if (
            all([c.derivation == Derivation.CONSTANT for c in self.output_columns])
            and not self.parent_ctes
            and not self.group_to_grain
        ):
            return False
        # if we don't need to source any concepts from anywhere
        # render without from
        # most likely to happen from inlining constants
        if not any([v for v in self.source_map.values()]):
            return False
        if (
            len(self.source.datasources) == 1
            and self.source.datasources[0].name == CONSTANT_DATASET
        ):
            return False
        return True

    @property
    def sourced_concepts(self) -> List[BuildConcept]:
        return [c for c in self.output_columns if c.address in self.source_map]


class ConceptPair(BaseModel):
    left: BuildConcept
    right: BuildConcept
    existing_datasource: Union[BuildDatasource, "QueryDatasource"]
    modifiers: List[Modifier] = Field(default_factory=list)

    @property
    def is_partial(self):
        return Modifier.PARTIAL in self.modifiers

    @property
    def is_nullable(self):
        return Modifier.NULLABLE in self.modifiers


class CTEConceptPair(ConceptPair):
    cte: CTE | UnionCTE


class InstantiatedUnnestJoin(BaseModel):
    object_to_unnest: BuildConcept | BuildParamaterizedConceptReference | BuildFunction
    alias: str = "unnest"


class UnnestJoin(BaseModel):
    concepts: list[BuildConcept]
    parent: BuildFunction
    alias: str = "unnest"
    rendering_required: bool = True

    def __hash__(self):
        return self.safe_identifier.__hash__()

    @property
    def safe_identifier(self) -> str:
        return self.alias + "".join([str(s.address) for s in self.concepts])


def raise_helpful_join_validation_error(
    concepts: List[BuildConcept],
    left_datasource: BuildDatasource | QueryDatasource | None,
    right_datasource: BuildDatasource | QueryDatasource | None,
):

    if not left_datasource or not right_datasource:
        raise InvalidSyntaxException(
            "No mutual keys found, and not two valid datasources"
        )
    left_keys = [c.address for c in left_datasource.output_concepts]
    right_keys = [c.address for c in right_datasource.output_concepts]
    match_concepts = [c.address for c in concepts]
    assert left_datasource
    assert right_datasource
    raise InvalidSyntaxException(
        "No mutual join keys found between"
        f" {left_datasource.identifier} and"
        f" {right_datasource.identifier}, left_keys {left_keys},"
        f" right_keys {right_keys},"
        f" provided join concepts {match_concepts}"
    )


class BaseJoin(BaseModel):
    right_datasource: Union[BuildDatasource, "QueryDatasource"]
    join_type: JoinType
    concepts: Optional[List[BuildConcept]] = None
    left_datasource: Optional[Union[BuildDatasource, "QueryDatasource"]] = None
    concept_pairs: list[ConceptPair] | None = None

    @model_validator(mode="after")
    def validate_join(self) -> "BaseJoin":
        if (
            self.left_datasource
            and self.left_datasource.identifier == self.right_datasource.identifier
        ):
            raise SyntaxError(
                f"Cannot join a dataself to itself, joining {self.left_datasource} and"
                f" {self.right_datasource}"
            )
        # Early returns maintained as in original code
        if self.concept_pairs or self.concepts == []:
            return self

        # reduce concept list to just the mutual keys
        final_concepts = []
        for concept in self.concepts or []:
            include = True
            for ds in [self.left_datasource, self.right_datasource]:
                synonyms = []
                if not ds:
                    continue
                for c in ds.output_concepts:
                    synonyms += list(c.pseudonyms)
                if (
                    concept.address not in ds.output_concepts
                    and concept.address not in synonyms
                ):
                    raise InvalidSyntaxException(
                        f"Invalid join, missing {concept} on {ds.name}, have"
                        f" {[c.address for c in ds.output_concepts]}"
                    )
            if include:
                final_concepts.append(concept)

        if not final_concepts and self.concepts:
            raise_helpful_join_validation_error(
                self.concepts,
                self.left_datasource,
                self.right_datasource,
            )

        self.concepts = final_concepts
        return self

    @property
    def unique_id(self) -> str:
        return str(self)

    @property
    def input_concepts(self) -> List[BuildConcept]:
        base = []
        if self.concept_pairs:
            for pair in self.concept_pairs:
                base += [pair.left, pair.right]
        elif self.concepts:
            base += self.concepts
        return base

    def __str__(self):
        if self.concept_pairs:
            return (
                f"{self.join_type.value} {self.right_datasource.name} on"
                f" {','.join([str(k.existing_datasource.name) + '.'+ str(k.left)+'='+str(k.right) for k in self.concept_pairs])}"
            )
        return (
            f"{self.join_type.value} {self.right_datasource.name} on"
            f" {','.join([str(k) for k in self.concepts])}"
        )


class QueryDatasource(BaseModel):
    input_concepts: List[BuildConcept]
    output_concepts: List[BuildConcept]
    datasources: List[Union[BuildDatasource, "QueryDatasource"]]
    source_map: Dict[str, Set[Union[BuildDatasource, "QueryDatasource", "UnnestJoin"]]]

    grain: BuildGrain
    joins: List[BaseJoin | UnnestJoin]
    limit: Optional[int] = None
    condition: Optional[
        Union[BuildConditional, BuildComparison, BuildParenthetical]
    ] = Field(default=None)
    source_type: SourceType = SourceType.SELECT
    partial_concepts: List[BuildConcept] = Field(default_factory=list)
    hidden_concepts: set[str] = Field(default_factory=set)
    nullable_concepts: List[BuildConcept] = Field(default_factory=list)
    join_derived_concepts: List[BuildConcept] = Field(default_factory=list)
    force_group: bool | None = None
    existence_source_map: Dict[str, Set[Union[BuildDatasource, "QueryDatasource"]]] = (
        Field(default_factory=dict)
    )
    ordering: BuildOrderBy | None = None

    def __repr__(self):
        return f"{self.identifier}@<{self.grain}>"

    @property
    def safe_identifier(self):
        return self.identifier.replace(".", "_")

    @property
    def full_concepts(self) -> List[BuildConcept]:
        return [
            c
            for c in self.output_concepts
            if c.address not in [z.address for z in self.partial_concepts]
        ]

    @field_validator("joins")
    @classmethod
    def validate_joins(cls, v):
        unique_pairs = set()
        for join in v:
            if not isinstance(join, BaseJoin):
                continue
            pairing = str(join)
            if pairing in unique_pairs:
                raise SyntaxError(f"Duplicate join {str(join)}")
            unique_pairs.add(pairing)
        return v

    @field_validator("input_concepts")
    @classmethod
    def validate_inputs(cls, v):
        return unique(v, "address")

    @field_validator("output_concepts")
    @classmethod
    def validate_outputs(cls, v):
        return unique(v, "address")

    @field_validator("source_map")
    @classmethod
    def validate_source_map(cls, v: dict, info: ValidationInfo):
        values = info.data
        hidden_concepts = values.get("hidden_concepts", set())
        for key in ("input_concepts", "output_concepts"):
            if not values.get(key):
                continue
            concept: BuildConcept
            for concept in values[key]:
                if concept.address in hidden_concepts:
                    continue
                if (
                    concept.address not in v
                    and not any(x in v for x in concept.pseudonyms)
                    and CONFIG.validate_missing
                ):
                    raise SyntaxError(
                        f"Missing source map entry for {concept.address} on {key} with pseudonyms {concept.pseudonyms}, have map: {v}"
                    )
        return v

    def __str__(self):
        return self.__repr__()

    def __hash__(self):
        return (self.identifier).__hash__()

    @property
    def concepts(self):
        return self.output_concepts

    @property
    def name(self):
        return self.identifier

    @property
    def group_required(self) -> bool:
        if self.force_group is True:
            return True
        if self.force_group is False:
            return False
        if self.source_type:
            if self.source_type in [
                SourceType.FILTER,
            ]:
                return False
            elif self.source_type in [
                SourceType.GROUP,
            ]:
                return True
        return False

    def __add__(self, other) -> "QueryDatasource":
        # these are syntax errors to avoid being caught by current
        if not isinstance(other, QueryDatasource):
            raise SyntaxError("Can only merge two query datasources")
        if not other.grain == self.grain:
            raise SyntaxError(
                "Can only merge two query datasources with identical grain"
            )
        if not self.group_required == other.group_required:
            raise SyntaxError(
                "can only merge two datasources if the group required flag is the same"
            )
        if not self.join_derived_concepts == other.join_derived_concepts:
            raise SyntaxError(
                "can only merge two datasources if the join derived concepts are the same"
            )
        if not self.force_group == other.force_group:
            raise SyntaxError(
                "can only merge two datasources if the force_group flag is the same"
            )
        logger.debug(
            f"[Query Datasource] merging {self.name} with"
            f" {[c.address for c in self.output_concepts]} concepts and"
            f" {other.name} with {[c.address for c in other.output_concepts]} concepts"
        )

        merged_datasources: dict[str, Union[BuildDatasource, "QueryDatasource"]] = {}

        for ds in [*self.datasources, *other.datasources]:
            if ds.safe_identifier in merged_datasources:
                merged_datasources[ds.safe_identifier] = (
                    merged_datasources[ds.safe_identifier] + ds
                )
            else:
                merged_datasources[ds.safe_identifier] = ds

        final_source_map: defaultdict[
            str, Set[Union[BuildDatasource, QueryDatasource, UnnestJoin]]
        ] = defaultdict(set)

        # add our sources
        for key in self.source_map:
            final_source_map[key] = self.source_map[key].union(
                other.source_map.get(key, set())
            )
        # add their sources
        for key in other.source_map:
            if key not in final_source_map:
                final_source_map[key] = other.source_map[key]

        # if a ds was merged (to combine columns), we need to update the source map
        # to use the merged item
        for k, v in final_source_map.items():
            final_source_map[k] = set(
                merged_datasources.get(x.safe_identifier, x) for x in list(v)
            )
        self_hidden: set[str] = self.hidden_concepts or set()
        other_hidden: set[str] = other.hidden_concepts or set()
        # hidden is the minimum overlapping set
        hidden = self_hidden.intersection(other_hidden)
        qds = QueryDatasource(
            input_concepts=unique(
                self.input_concepts + other.input_concepts, "address"
            ),
            output_concepts=unique(
                self.output_concepts + other.output_concepts, "address"
            ),
            source_map=final_source_map,
            datasources=list(merged_datasources.values()),
            grain=self.grain,
            joins=unique(self.joins + other.joins, "unique_id"),
            condition=(
                self.condition + other.condition
                if self.condition and other.condition
                else self.condition or other.condition
            ),
            source_type=self.source_type,
            partial_concepts=unique(
                self.partial_concepts + other.partial_concepts, "address"
            ),
            join_derived_concepts=self.join_derived_concepts,
            force_group=self.force_group,
            hidden_concepts=hidden,
            ordering=self.ordering,
        )
        logger.debug(
            f"[Query Datasource] merged with {[c.address for c in qds.output_concepts]} concepts"
        )
        logger.debug(qds.source_map)
        return qds

    @property
    def identifier(self) -> str:
        filters = abs(hash(str(self.condition))) if self.condition else ""
        grain = "_".join([str(c).replace(".", "_") for c in self.grain.components])
        group = ""
        if self.group_required:
            keys = [
                x.address for x in self.output_concepts if x.purpose != Purpose.METRIC
            ]
            group = "_grouped_by_" + "_".join(keys)
        return (
            "_join_".join([d.identifier for d in self.datasources])
            + group
            + (f"_at_{grain}" if grain else "_at_abstract")
            + (f"_filtered_by_{filters}" if filters else "")
        )

    def get_alias(
        self,
        concept: BuildConcept,
        use_raw_name: bool = False,
        force_alias: bool = False,
        source: str | None = None,
    ):
        for x in self.datasources:
            # query datasources should be referenced by their alias, always
            force_alias = isinstance(x, QueryDatasource)
            #
            use_raw_name = isinstance(x, DATASOURCE_TYPES) and not force_alias
            if source and x.safe_identifier != source:
                continue
            try:
                return x.get_alias(
                    concept.with_grain(self.grain),
                    use_raw_name,
                    force_alias=force_alias,
                )
            except ValueError:
                continue
        existing = [c.with_grain(self.grain) for c in self.output_concepts]
        if concept in existing:
            return concept.name

        existing_str = [str(c) for c in existing]
        datasources = [ds.identifier for ds in self.datasources]
        raise ValueError(
            f"{LOGGER_PREFIX} Concept {str(concept)} not found on {self.identifier};"
            f" have {existing_str} from {datasources}."
        )

    @property
    def safe_location(self):
        return self.identifier


class AliasedExpression(BaseModel):
    expr: BuildExpr
    alias: str


class RecursiveCTE(CTE):

    def generate_loop_functions(
        self,
        recursive_derived: BuildConcept,
        left_recurse_concept: BuildConcept,
        right_recurse_concept: BuildConcept,
    ) -> tuple[BuildConcept, BuildConcept, BuildConcept]:

        join_gate = BuildConcept(
            name=RECURSIVE_GATING_CONCEPT,
            namespace=DEFAULT_NAMESPACE,
            grain=recursive_derived.grain,
            build_is_aggregate=False,
            datatype=DataType.BOOL,
            purpose=Purpose.KEY,
            derivation=Derivation.BASIC,
            lineage=BuildFunction(
                operator=FunctionType.CASE,
                arguments=[
                    BuildCaseWhen(
                        comparison=BuildComparison(
                            left=right_recurse_concept,
                            right=MagicConstants.NULL,
                            operator=ComparisonOperator.IS,
                        ),
                        expr=True,
                    ),
                    BuildCaseElse(expr=False),
                ],
                output_data_type=DataType.BOOL,
                output_purpose=Purpose.KEY,
            ),
        )
        bottom_join_gate = BuildConcept(
            name=f"{RECURSIVE_GATING_CONCEPT}_two",
            namespace=DEFAULT_NAMESPACE,
            grain=recursive_derived.grain,
            build_is_aggregate=False,
            datatype=DataType.BOOL,
            purpose=Purpose.KEY,
            derivation=Derivation.BASIC,
            lineage=BuildFunction(
                operator=FunctionType.CASE,
                arguments=[
                    BuildCaseWhen(
                        comparison=BuildComparison(
                            left=right_recurse_concept,
                            right=MagicConstants.NULL,
                            operator=ComparisonOperator.IS,
                        ),
                        expr=True,
                    ),
                    BuildCaseElse(expr=False),
                ],
                output_data_type=DataType.BOOL,
                output_purpose=Purpose.KEY,
            ),
        )
        bottom_derivation = BuildConcept(
            name=recursive_derived.name + "_bottom",
            namespace=recursive_derived.namespace,
            grain=recursive_derived.grain,
            build_is_aggregate=False,
            datatype=recursive_derived.datatype,
            purpose=recursive_derived.purpose,
            derivation=Derivation.RECURSIVE,
            lineage=BuildFunction(
                operator=FunctionType.CASE,
                arguments=[
                    BuildCaseWhen(
                        comparison=BuildComparison(
                            left=right_recurse_concept,
                            right=MagicConstants.NULL,
                            operator=ComparisonOperator.IS,
                        ),
                        expr=recursive_derived,
                    ),
                    BuildCaseElse(expr=right_recurse_concept),
                ],
                output_data_type=recursive_derived.datatype,
                output_purpose=recursive_derived.purpose,
            ),
        )
        return bottom_derivation, join_gate, bottom_join_gate

    @property
    def internal_ctes(self) -> List[CTE]:
        filtered_output = [
            x for x in self.output_columns if x.name != RECURSIVE_GATING_CONCEPT
        ]
        recursive_derived = [
            x for x in self.output_columns if x.derivation == Derivation.RECURSIVE
        ][0]
        if not isinstance(recursive_derived.lineage, BuildFunction):
            raise SyntaxError(
                "Recursive CTEs must have a function lineage, found"
                f" {recursive_derived.lineage}"
            )
        left_recurse_concept = recursive_derived.lineage.concept_arguments[0]
        right_recurse_concept = recursive_derived.lineage.concept_arguments[1]
        parent_ctes: List[CTE | UnionCTE]
        if self.parent_ctes:
            base = self.parent_ctes[0]
            loop_input_cte = base
            parent_ctes = [base]
            parent_identifier = base.identifier
        else:
            raise SyntaxError("Recursive CTEs must have a parent CTE currently")
        bottom_derivation, join_gate, bottom_join_gate = self.generate_loop_functions(
            recursive_derived, left_recurse_concept, right_recurse_concept
        )
        base_output = [*filtered_output, join_gate]
        bottom_output = []
        for x in filtered_output:
            if x.address == recursive_derived.address:
                bottom_output.append(bottom_derivation)
            else:
                bottom_output.append(x)

        bottom_output = [*bottom_output, bottom_join_gate]
        top = CTE(
            name=self.name,
            source=self.source,
            output_columns=base_output,
            source_map=self.source_map,
            grain=self.grain,
            existence_source_map=self.existence_source_map,
            parent_ctes=self.parent_ctes,
            joins=self.joins,
            condition=self.condition,
            partial_concepts=self.partial_concepts,
            hidden_concepts=self.hidden_concepts,
            nullable_concepts=self.nullable_concepts,
            join_derived_concepts=self.join_derived_concepts,
            group_to_grain=self.group_to_grain,
            order_by=self.order_by,
            limit=self.limit,
        )
        top_cte_array: list[CTE | UnionCTE] = [top]
        bottom_source_map = {
            left_recurse_concept.address: [top.identifier],
            right_recurse_concept.address: [parent_identifier],
            # recursive_derived.address: self.source_map[recursive_derived.address],
            join_gate.address: [top.identifier],
            recursive_derived.address: [top.identifier],
        }
        bottom = CTE(
            name=self.name,
            source=self.source,
            output_columns=bottom_output,
            source_map=bottom_source_map,
            grain=self.grain,
            existence_source_map=self.existence_source_map,
            parent_ctes=top_cte_array + parent_ctes,
            joins=[
                Join(
                    right_cte=loop_input_cte,
                    jointype=JoinType.INNER,
                    joinkey_pairs=[
                        CTEConceptPair(
                            left=recursive_derived,
                            right=left_recurse_concept,
                            existing_datasource=loop_input_cte.source,
                            modifiers=[],
                            cte=top,
                        )
                    ],
                    condition=BuildComparison(
                        left=join_gate, right=True, operator=ComparisonOperator.IS_NOT
                    ),
                )
            ],
            partial_concepts=self.partial_concepts,
            hidden_concepts=self.hidden_concepts,
            nullable_concepts=self.nullable_concepts,
            join_derived_concepts=self.join_derived_concepts,
            group_to_grain=self.group_to_grain,
            order_by=self.order_by,
            limit=self.limit,
        )
        return [top, bottom]


class UnionCTE(BaseModel):
    name: str
    source: QueryDatasource
    parent_ctes: list[CTE | UnionCTE]
    internal_ctes: list[CTE | UnionCTE]
    output_columns: List[BuildConcept]
    grain: BuildGrain
    operator: str = "UNION ALL"
    order_by: Optional[BuildOrderBy] = None
    limit: Optional[int] = None
    hidden_concepts: set[str] = Field(default_factory=set)
    partial_concepts: list[BuildConcept] = Field(default_factory=list)
    existence_source_map: Dict[str, list[str]] = Field(default_factory=dict)
    inlined_ctes: Dict[str, InlinedCTE] = Field(default_factory=dict)

    @computed_field  # type: ignore
    @property
    def output_lcl(self) -> LooseBuildConceptList:
        return LooseBuildConceptList(concepts=self.output_columns)

    def get_alias(self, concept: BuildConcept, source: str | None = None) -> str:
        for cte in self.parent_ctes:
            if concept.address in cte.output_columns:
                if source and source != cte.name:
                    continue
                return concept.safe_address
        return "INVALID_ALIAS"

    def get_concept(self, address: str) -> BuildConcept | None:
        for cte in self.internal_ctes:
            if address in cte.output_columns:
                match = [x for x in cte.output_columns if x.address == address].pop()
                return match

        match_list = [x for x in self.output_columns if x.address == address]
        if match_list:
            return match_list.pop()
        return None

    @property
    def source_map(self):
        return {x.address: [] for x in self.output_columns}

    @property
    def condition(self):
        return None

    @condition.setter
    def condition(self, value):
        raise NotImplementedError

    @property
    def identifier(self) -> str:
        return self.name

    @property
    def safe_identifier(self):
        return self.name

    @property
    def group_to_grain(self) -> bool:
        return False

    @property
    def group_concepts(self) -> List[BuildConcept]:
        # unions should always be on unique sets
        return []

    def __add__(self, other):
        if not isinstance(other, UnionCTE) or not other.name == self.name:
            raise SyntaxError("Cannot merge union CTEs")
        return self


class Join(BaseModel):
    right_cte: CTE | UnionCTE
    jointype: JoinType
    left_cte: CTE | UnionCTE | None = None
    joinkey_pairs: List[CTEConceptPair] | None = None
    inlined_ctes: set[str] = Field(default_factory=set)
    quote: str | None = None
    condition: BuildConditional | BuildComparison | BuildParenthetical | None = None

    def inline_cte(self, cte: CTE):
        self.inlined_ctes.add(cte.name)

    def get_name(self, cte: CTE | UnionCTE) -> str:
        if cte.identifier in self.inlined_ctes:
            return cte.source.datasources[0].safe_identifier
        return cte.safe_identifier

    @property
    def right_name(self) -> str:
        if self.right_cte.identifier in self.inlined_ctes:
            return self.right_cte.source.datasources[0].safe_identifier
        return self.right_cte.safe_identifier

    @property
    def right_ref(self) -> str:
        if self.quote:
            if self.right_cte.identifier in self.inlined_ctes:
                return f"{safe_quote(self.right_cte.source.datasources[0].safe_location, self.quote)} as {self.quote}{self.right_cte.source.datasources[0].safe_identifier}{self.quote}"
            return f"{self.quote}{self.right_cte.safe_identifier}{self.quote}"
        if self.right_cte.identifier in self.inlined_ctes:
            return f"{self.right_cte.source.datasources[0].safe_location} as {self.right_cte.source.datasources[0].safe_identifier}"
        return self.right_cte.safe_identifier

    @property
    def unique_id(self) -> str:
        return str(self)

    def __str__(self):
        if self.joinkey_pairs:
            return (
                f"{self.jointype.value} join"
                f" {self.right_name} on"
                f" {','.join([k.cte.name + '.'+str(k.left.address)+'='+str(k.right.address) for k in self.joinkey_pairs])}"
            )
        elif self.left_cte:
            return (
                f"{self.jointype.value} JOIN {self.left_cte.name} and"
                f" {self.right_name} on {','.join([str(k) for k in self.joinkey_pairs])}"
            )
        return f"{self.jointype.value} JOIN  {self.right_name} on {','.join([str(k) for k in self.joinkey_pairs])}"


def merge_ctes(ctes: List[CTE | UnionCTE]) -> List[CTE | UnionCTE]:
    final_ctes_dict: Dict[str, CTE | UnionCTE] = {}
    # merge CTEs
    for cte in ctes:
        if cte.name not in final_ctes_dict:
            final_ctes_dict[cte.name] = cte
        else:
            final_ctes_dict[cte.name] = final_ctes_dict[cte.name] + cte

    final_ctes = list(final_ctes_dict.values())
    return final_ctes


class CompiledCTE(BaseModel):
    name: str
    statement: str


UnionCTE.model_rebuild()



from typing import List

from trilogy.core.enums import SourceType
from trilogy.core.models.build import BuildConcept
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.models.execute import QueryDatasource
from trilogy.core.processing.nodes.base_node import StrategyNode


class RecursiveNode(StrategyNode):
    """Union nodes represent combining two keyspaces"""

    source_type = SourceType.RECURSIVE

    def __init__(
        self,
        input_concepts: List[BuildConcept],
        output_concepts: List[BuildConcept],
        environment: BuildEnvironment,
        whole_grain: bool = False,
        parents: List["StrategyNode"] | None = None,
        depth: int = 0,
    ):
        super().__init__(
            input_concepts=input_concepts,
            output_concepts=output_concepts,
            environment=environment,
            whole_grain=whole_grain,
            parents=parents,
            depth=depth,
        )

    def _resolve(self) -> QueryDatasource:
        """We need to ensure that any filtered values are removed from the output to avoid inappropriate references"""
        base = super()._resolve()
        return base

    def copy(self) -> "RecursiveNode":
        return RecursiveNode(
            input_concepts=list(self.input_concepts),
            output_concepts=list(self.output_concepts),
            environment=self.environment,
            whole_grain=self.whole_grain,
            parents=self.parents,
            depth=self.depth,
        )
